""" Library for building Dataset. """
from typing import List, Dict, Any
from copy import copy
import itertools

from scipy.spatial.distance import pdist, squareform
import numpy as npy
from sklearn import preprocessing

try:
    from plot_data.core import Scatter, Histogram, MultiplePlots, Tooltip, ParallelPlot, PointFamily, EdgeStyle, Axis, \
        PointStyle, Sample
    from plot_data.colors import BLUE, GREY
except ImportError:
    pass
from dessia_common.core import DessiaObject, DessiaFilter, FiltersList
from dessia_common.exports import MarkdownWriter
from dessia_common import templates
from dessia_common.datatools.metrics import mean, std, variance, covariance_matrix


class Dataset(DessiaObject):
    """
    Base object for handling a list of DessiaObjects.

    :param dessia_objects:
        --------
        List of DessiaObjects to store in Dataset
    :type dessia_objects: List[DessiaObject], `optional`, defaults to `None`

    :param name:
        --------
        Name of Dataset
    :type name: str, `optional`, defaults to `''`

    :Properties:
        * **common_attributes:** (`List[str]`)
            --------
            Common attributes of DessiaObjects contained in the current `Dataset`

        * **matrix:** (`List[List[float]]`, `n_samples x n_features`)
            --------
            Matrix of data computed by calling the `to_vector` method of all `dessia_objects`

    **Built-in methods**:
        * __init__
            >>> from dessia_common.datatools.dataset import Dataset
            >>> from dessia_common.models import all_cars_wi_feat
            >>> hlist = Dataset(all_cars_wi_feat, name="init")

        * __str__
            >>> print(Dataset(all_cars_wi_feat[:3], name='printed'))
            Dataset printed: 3 samples, 5 features
            |         Mpg         |    Displacement    |     Horsepower     |       Weight       |    Acceleration    |
            -----------------------------------------------------------------------------------------------------------
            |               18.0  |             0.307  |             130.0  |            3504.0  |              12.0  |
            |               15.0  |              0.35  |             165.0  |            3693.0  |              11.5  |
            |               18.0  |             0.318  |             150.0  |            3436.0  |              11.0  |

        * __len__
            >>> len(Dataset(all_cars_wi_feat))
            returns len(all_cars_wi_feat)

        * __get_item__
            >>> Dataset(all_cars_wi_feat)[0]
            returns <dessia_common.tests.CarWithFeatures object at 'memory_address'>
            >>> Dataset(all_cars_wi_feat)[0:2]
            returns Dataset(all_cars_wi_feat[0:2])
            >>> Dataset(all_cars_wi_feat)[[0,5,6]]
            returns Dataset([all_cars_wi_feat[i] for i in [0,5,6]])
            >>> booleans_list = [True, False,..., True] of length len(all_cars_wi_feat)
            >>> Dataset(all_cars_wi_feat)[booleans_list]
            returns Dataset([car for car, boolean in zip(all_cars_wi_feat, booleans_list) if boolean])

        * __add__
            >>> Dataset(all_cars_wi_feat) + Dataset(all_cars_wi_feat)
            Dataset(all_cars_wi_feat + all_cars_wi_feat)
            >>> Dataset(all_cars_wi_feat) + Dataset()
            Dataset(all_cars_wi_feat)
            >>> Dataset(all_cars_wi_feat).extend(Dataset(all_cars_wi_feat))
            Dataset(all_cars_wi_feat + all_cars_wi_feat)
    """

    _standalone_in_db = True
    _vector_features = ["name", "common_attributes"]
    _non_data_eq_attributes = ["name", "_common_attributes", "_matrix"]

    def __init__(self, dessia_objects: List[DessiaObject] = None, name: str = ''):
        """ See class docstring. """
        if dessia_objects is None:
            dessia_objects = []
        self.dessia_objects = dessia_objects
        self._common_attributes = None
        self._matrix = None
        DessiaObject.__init__(self, name=name)

    def __getitem__(self, key: Any):
        """
        Custom Dataset function to get item.

        In addition to work as numpy.arrays of dimension `(n,)`, allows to pick a sub-Dataset from a list of indexes.
        """
        if len(self) == 0:
            return []
        if isinstance(key, int):
            return self._pick_from_int(key)
        if isinstance(key, slice):
            return self._pick_from_slice(key)
        if isinstance(key, list):
            if len(key) == 0:
                return self.__class__()
            if isinstance(key[0], bool):
                if len(key) == self.__len__():
                    return self._pick_from_boolist(key)
                raise ValueError(f"Cannot index {self.__class__.__name__} object of len {self.__len__()} with a "
                                 f"list of boolean of len {len(key)}")
            if isinstance(key[0], int):
                return self._pick_from_boolist(self._indexlist_to_booleanlist(key))

            raise NotImplementedError(f"key of type {type(key)} with {type(key[0])} elements not implemented for "
                                      f"indexing Datasets")

        raise NotImplementedError(f"key of type {type(key)} not implemented for indexing Datasets")

    def __add__(self, other: 'Dataset'):
        """ Allows to merge two Dataset into one by merging their dessia_object into one list. """
        if self.__class__ != Dataset or other.__class__ != Dataset:
            raise TypeError("Addition only defined for Dataset. A specific __add__ method is required for "
                            f"{self.__class__}")

        sum_hlist = self.__class__(dessia_objects=self.dessia_objects + other.dessia_objects,
                                   name=self.name[:5] + '_+_' + other.name[:5])

        if all(item in self.common_attributes for item in other.common_attributes):
            sum_hlist._common_attributes = self.common_attributes
            if self._matrix is not None and other._matrix is not None:
                sum_hlist._matrix = self._matrix + other._matrix
        return sum_hlist

    def extend(self, other: 'Dataset'):
        """
        Update a Dataset by adding b values to it.

        :param b: Dataset to add to the current Dataset
        :type b: Dataset

        :return: None

        :Examples:
        >>> from dessia_common.datatools.dataset import Dataset
        >>> from dessia_common.models import all_cars_wi_feat
        >>> Dataset(all_cars_wi_feat).extend(Dataset(all_cars_wi_feat))
        Dataset(all_cars_wi_feat + all_cars_wi_feat)
        """
        # Not "self.dessia_objects += other.dessia_objects" to take advantage of __add__ algorithm
        self.__dict__.update((self + other).__dict__)

    def _pick_from_int(self, idx: int):
        return self.dessia_objects[idx]

    def _pick_from_slice(self, key: slice):
        new_hlist = self.__class__(dessia_objects=self.dessia_objects[key], name=self.name)
        new_hlist._common_attributes = copy(self._common_attributes)
        new_hlist.dessia_objects = self.dessia_objects[key]
        if self._matrix is not None:
            new_hlist._matrix = self._matrix[key]
        # new_hlist.name += f"_{key.start if key.start is not None else 0}_{key.stop}")
        return new_hlist

    def _indexlist_to_booleanlist(self, index_list: List[int]):
        boolean_list = [False] * len(self)
        for idx in index_list:
            boolean_list[idx] = True
        return boolean_list

    def _pick_from_boolist(self, key: List[bool]):
        new_hlist = self.__class__(dessia_objects=DessiaFilter.apply(self.dessia_objects, key), name=self.name)
        new_hlist._common_attributes = copy(self._common_attributes)
        if self._matrix is not None:
            new_hlist._matrix = DessiaFilter.apply(self._matrix, key)
        # new_hlist.name += "_list")
        return new_hlist

    def __str__(self):
        """ Print Dataset as a table. """
        attr_space = []

        prefix = self._write_str_prefix()

        if self.__len__() == 0:
            return prefix

        string = ""
        string += self._print_titles(attr_space)
        string += "\n" + "-" * len(string)
        string += self._print_objects_slice(slice(0, 5), attr_space)

        if len(self) > 10:
            undispl_len = len(self) - 10
            string += (f"\n+ {undispl_len} undisplayed object" + "s" * (min([undispl_len, 2]) - 1) + "...")

        if len(self) > 5:
            string += self._print_objects_slice(slice(-5, len(self)), attr_space)
        return prefix + "\n" + string + "\n"

    def _printed_attributes(self):
        if 'name' in self.common_attributes:
            return self.common_attributes
        return ['name'] + self.common_attributes

    def _print_objects_slice(self, key: slice, attr_space: List[int]):
        string = ""
        for index in range(len(self[key])):
            string += "\n"
            string += self._print_object(index, attr_space)
        return string

    def _write_str_prefix(self):
        prefix = f"{self.__class__.__name__} {self.name if self.name != '' else hex(id(self))}: "
        prefix += f"{len(self)} samples, {len(self.common_attributes)} features"
        return prefix

    def _print_titles(self, attr_space: List[int]):
        min_col_length = 16
        printed_attributes = self._printed_attributes()
        string = ""
        for idx, attr in enumerate(printed_attributes):
            end_bar = ""
            if idx == len(printed_attributes) - 1:
                end_bar = "|"
            # attribute
            if len(attr) + 6 > min_col_length:
                indentation = 3
            else:
                indentation = min_col_length - len(attr)
                odd_incr = int(indentation % 2)
                indentation = int(indentation / 2)

            name_attr = " " * indentation + " " * odd_incr + f"{attr.capitalize()}" + " " * indentation
            attr_space.append(len(name_attr))
            string += "|" + name_attr + end_bar
        return string

    def _print_object(self, index: int, attr_space: List[int]):
        printed_attributes = self._printed_attributes()
        string = ""
        for idx, attr in enumerate(printed_attributes):
            end_bar = ""
            if idx == len(printed_attributes) - 1:
                end_bar = "|"

            attr_value = self._get_printed_value(index, attr)

            string += "|" + " " * max((attr_space[idx] - len(str(attr_value)) - 1), 1)
            string += f"{attr_value}"[:attr_space[idx] - 4]
            if len(str(attr_value)) > attr_space[idx] - 3:
                string = string[:-1] + "... "
            elif len(str(attr_value)) == attr_space[idx] - 3:
                string += f"{attr_value}"[-1] + " "
            else:
                string += " "
            string += end_bar
        return string

    def to_markdown(self, *args, **kwargs) -> str:
        """Render a markdown of the object output type: string."""
        md_writer = MarkdownWriter(print_limit=25, table_limit=12)
        name = md_writer.print_name(self)
        class_ = md_writer.print_class(self)
        element_details = md_writer.element_details(self)
        table = md_writer.matrix_table(self.matrix, self.common_attributes)

        return templates.dataset_markdown_template.substitute(name=name, class_=class_, element_details=element_details,
                                                              table=table)

    def _get_printed_value(self, index: int, attr: str):
        try:
            return getattr(self[index], attr)
        except AttributeError:
            return self.matrix[index][self.common_attributes.index(attr)]

    def __len__(self):
        """Length of Dataset is len(Dataset.dessia_objects)."""
        return len(self.dessia_objects)

    @property
    def common_attributes(self):
        """List of common attributes of stored dessia_objects."""
        if self._common_attributes is None:
            if len(self) == 0:
                return []

            all_class = []
            one_instance = []
            for dessia_object in self.dessia_objects:
                if dessia_object.__class__ not in all_class:
                    all_class.append(dessia_object.__class__)
                    one_instance.append(dessia_object)

            all_attributes = sum((instance.vector_features() for instance in one_instance), [])
            set_attributes = set.intersection(*(set(instance.vector_features()) for instance in one_instance))

            # Keep order
            self._common_attributes = []
            for attr in all_attributes:
                if attr in set_attributes:
                    self._common_attributes.append(attr)
                    set_attributes.remove(attr)

        return self._common_attributes

    @property
    def matrix(self):
        """
        Get equivalent matrix of dessia_objects.

        Dimensions: `len(dessia_objects) x len(common_attributes)`.
        """
        if self._matrix is None:
            matrix = []
            for dessia_object in self.dessia_objects:
                temp_row = dessia_object.to_vector()
                vector_features = dessia_object.vector_features()
                matrix.append(list(temp_row[vector_features.index(attr)] for attr in self.common_attributes))
            self._matrix = matrix
        return self._matrix

    def attribute_values(self, attribute: str):
        """
        Get a list of all values of dessia_objects of an attribute given by name.

        :param attribute: Attribute to get all values
        :type attribute: str

        :return: A list of all values of the specified attribute of dessia_objects
        :rtype: List[Any]

        :Examples:
        >>> from dessia_common.datatools.dataset import Dataset
        >>> from dessia_common.models import all_cars_wi_feat
        >>> Dataset(all_cars_wi_feat[:10]).attribute_values("weight")
        [3504.0, 3693.0, 3436.0, 3433.0, 3449.0, 4341.0, 4354.0, 4312.0, 4425.0, 3850.0]
        """
        if not hasattr(self.dessia_objects[0], attribute):
            if attribute not in self.common_attributes:
                raise ValueError(f"{attribute} not in common_attributes = {self.common_attributes}")
            return self.column_values(self.common_attributes.index(attribute))
        return [getattr(dessia_object, attribute) for dessia_object in self.dessia_objects]

    def column_values(self, index: int):
        """
        Get a list of all values of dessia_objects for an attribute given by its index common_attributes.

        :param index: Index in common_attributes to get all values of dessia_objects
        :type index: int

        :return: A list of all values of the specified attribute of dessia_objects
        :rtype: List[float]

        :Examples:
        >>> from dessia_common.datatools.dataset import Dataset
        >>> from dessia_common.models import all_cars_wi_feat
        >>> Dataset(all_cars_wi_feat[:10]).column_values(2)
        [130.0, 165.0, 150.0, 150.0, 140.0, 198.0, 220.0, 215.0, 225.0, 190.0]
        """
        return [row[index] for row in self.matrix]

    def sub_matrix(self, columns_names: List[str]):
        """
        Build a sub matrix of the current Dataset taking column numbers in indexes or attribute values in attributes.

        Warning: Only one of `indexes` or `attributes` has to be specified.

        :param columns_names: List of columns' names to create a sub matrix
        :type columns_names: List[str]

        :return: Data stored in matrix reduced to the specified `indexes` or `attributes`
        :rtype: List[List[float]]

        :Examples:
        >>> from dessia_common.datatools.dataset import Dataset
        >>> from dessia_common.models import all_cars_wi_feat
        >>> print(Dataset(all_cars_wi_feat[:10]).sub_matrix(['displacement', 'horsepower']))
        [[0.307, 130.0], [0.35, 165.0], [0.318, 150.0], [0.304, 150.0], [0.302, 140.0], [0.429, 198.0],
         [0.454, 220.0], [0.44, 215.0], [0.455, 225.0], [0.39, 190.0]]
        """
        transposed_submatrix = [self.attribute_values(column_name) for column_name in columns_names]
        return list(map(list, zip(*transposed_submatrix)))

    def sort(self, key: Any, ascend: bool = True):  # TODO : Replace numpy with faster algorithms
        """
        Sort the current Dataset along the given key.

        :param key:
            --------
            The parameter on which to sort the Dataset. Can be an attribute or its index in \
                `common_attributes`
        :type key: int or str

        :param ascend:
            --------
            Whether to sort the Dataset in ascending (`True`) or descending (`False`) order
        :type key: bool, defaults to `True`

        :return: None

        :Examples:
        >>> from dessia_common.datatools.dataset import Dataset
        >>> from dessia_common.models import all_cars_wi_feat
        >>> example_list = Dataset(all_cars_wi_feat[:3], "sort_example")
        >>> example_list.sort("mpg", False)
        >>> print(example_list)
        Dataset sort_example: 3 samples, 5 features
        |         Mpg         |    Displacement    |     Horsepower     |       Weight       |    Acceleration    |
        -----------------------------------------------------------------------------------------------------------
        |               18.0  |             0.318  |             150.0  |            3436.0  |              11.0  |
        |               18.0  |             0.307  |             130.0  |            3504.0  |              12.0  |
        |               15.0  |              0.35  |             165.0  |            3693.0  |              11.5  |
        >>> example_list.sort(2, True)
        >>> print(example_list)
        Dataset sort_example: 3 samples, 5 features
        |         Mpg         |    Displacement    |     Horsepower     |       Weight       |    Acceleration    |
        -----------------------------------------------------------------------------------------------------------
        |               18.0  |             0.307  |             130.0  |            3504.0  |              12.0  |
        |               18.0  |             0.318  |             150.0  |            3436.0  |              11.0  |
        |               15.0  |              0.35  |             165.0  |            3693.0  |              11.5  |
        """
        if len(self) != 0:
            if isinstance(key, int):
                sort_indexes = npy.argsort(self.column_values(key))
            elif isinstance(key, str):
                sort_indexes = npy.argsort(self.attribute_values(key))
            self.dessia_objects = [self.dessia_objects[idx] for idx in (sort_indexes if ascend else sort_indexes[::-1])]
            if self._matrix is not None:
                self._matrix = [self._matrix[idx] for idx in
                                (sort_indexes if ascend else sort_indexes[::-1])]

    def mean(self):
        """
        Compute means along each `common_attribute`.

        :return: A list of means along each dimension
        :rtype: List[float]

        :Examples:
        >>> from dessia_common.datatools.dataset import Dataset
        >>> from dessia_common.models import all_cars_wi_feat
        >>> example_list = Dataset(all_cars_wi_feat, "mean_example")
        >>> print(example_list.mean())
        [23.051231527093602, 0.1947795566502462, 103.5295566502463, 2979.4137931034484, 15.519704433497521]
        """
        return [mean(row) for row in zip(*self.matrix)]

    def standard_deviation(self):
        """
        Compute standard deviations along each `common_attribute`.

        :return: A list of standard deviations along each dimension
        :rtype: List[float]

        :Examples:
        >>> from dessia_common.datatools.dataset import Dataset
        >>> from dessia_common.models import all_cars_wi_feat
        >>> example_list = Dataset(all_cars_wi_feat, "std_example")
        >>> print(example_list.standard_deviation())
        [8.391423956652817, 0.10479316386533469, 40.47072606559397, 845.9605763601298, 2.799904275515381]
        """
        return [std(row) for row in zip(*self.matrix)]

    def variances(self):
        """
        Compute variances along each `common_attribute`.

        :return: A list of variances along each dimension
        :rtype: List[float]

        :Examples:
        >>> from dessia_common.datatools.dataset import Dataset
        >>> from dessia_common.models import all_cars_wi_feat
        >>> example_list = Dataset(all_cars_wi_feat, "var_example")
        >>> print(example_list.variances())
        [70.41599602028683, 0.010981607192906888, 1637.8796682763475, 715649.2967555631, 7.839463952049309]
        """
        return [variance(row) for row in zip(*self.matrix)]

    def covariance_matrix(self):
        """
        Compute the covariance matrix of `self.matrix`.

        :return: the covariance matrix of all stored data in self
        :rtype: List[List[float]], `n_features x n_features`

        :Examples:
        >>> from dessia_common.datatools.dataset import Dataset
        >>> from dessia_common.models import all_cars_wi_feat
        >>> example_list = Dataset(all_cars_wi_feat, "covar_example")
        >>> cov_matrix = example_list.covariance_matrix()
        >>> for row in cov_matrix: print(row)
        [70.58986267712706, -0.6737370735267286, -247.39164142796338, -5604.189893571734, 9.998099130329008]
        [-0.6737370735267286, 0.011008722272395539, 3.714807148938756, 82.86881366538952, -0.16412268260049875]
        [-247.39164142796338, 3.714807148938756, 1641.9238156054248, 28857.60749255002, -77.47638630420236]
        [-5604.189893571734, 82.86881366538952, 28857.60749255002, 717416.332056194, -1021.2202724563649]
        [9.998099130329008, -0.16412268260049875, -77.47638630420236, -1021.2202724563649, 7.8588206531654805]
        """
        return covariance_matrix(list(zip(*self.matrix)))

    def distance_matrix(self, method: str = 'minkowski', **kwargs):
        """
        Compute the distance matrix of `self.matrix`, i.e. the pairwise distances between all dessia_objects.

        Distances are computed with numerical values of `self.matrix`.

        :param method:
            Method to compute distances.
            Can be one of `[‘braycurtis’, ‘canberra’, ‘chebyshev’, ‘cityblock’, ‘correlation’, ‘cosine’, ‘dice’, \
            ‘euclidean’, ‘hamming’, ‘jaccard’, ‘jensenshannon’, ‘kulczynski1’, ‘mahalanobis’, ‘matching’, ‘minkowski’, \
            ‘rogerstanimoto’, ‘russellrao’, ‘seuclidean’, ‘sokalmichener’, ‘sokalsneath’, ‘sqeuclidean’, ‘yule’]`.
        :type method: str, `optional`, defaults to `'minkowski'`

        :param **kwargs:
            |  Extra arguments to metric: refer to each metric documentation for a list of all possible arguments.
            |  Some possible arguments:
            |     - p : scalar The p-norm to apply for Minkowski, weighted and unweighted. Default: `2`.
            |     - w : array_like The weight vector for metrics that support weights (e.g., Minkowski).
            |     - V : array_like The variance vector for standardized Euclidean. Default: \
                `var(vstack([XA, XB]), axis=0, ddof=1)`
            |     - VI : array_like The inverse of the covariance matrix for Mahalanobis. Default: \
                `inv(cov(vstack([XA, XB].T))).T`
            |     - out : ndarray The output array If not None, the distance matrix Y is stored in this array.
        :type **kwargs: dict, `optional`

        :return: the distance matrix of all stored data in self
        :rtype: List[List[float]], `n_samples x n_samples`

        :Examples:
        >>> from dessia_common.datatools.dataset import Dataset
        >>> from dessia_common.models import all_cars_wi_feat
        >>> example_list = Dataset(all_cars_wi_feat, "distance_example")
        >>> distance_matrix = example_list.distance_matrix('mahalanobis')
        >>> for row in distance_matrix[:4]: print(row[:4])
        [0.0, 1.6150355142162274, 1.0996902429379676, 1.3991408510938068]
        [1.6150355142162274, 0.0, 0.7691239946132247, 0.6216479207905371]
        [1.0996902429379676, 0.7691239946132247, 0.0, 0.7334135920381655]
        [1.3991408510938068, 0.6216479207905371, 0.7334135920381655, 0.0]
        """
        kwargs = self._set_distance_kwargs(method, kwargs)
        distances = squareform(pdist(self.matrix, method, **kwargs)).astype(float)
        return distances.tolist()

    @staticmethod
    def _set_distance_kwargs(method: str, kwargs: Dict[str, Any]):
        if 'p' not in kwargs and method == 'minkowski':
            kwargs['p'] = 2
        return kwargs

    def filtering(self, filters_list: FiltersList):
        """
        Filter a Dataset given a FiltersList. Method filtering apply a FiltersList to the current Dataset.

        :param filters_list:
            FiltersList to apply on current Dataset
        :type filters_list: FiltersList

        :return: The filtered Dataset
        :rtype: Dataset

        :Examples:
        >>> from dessia_common.core import DessiaFilter
        >>> from dessia_common.datatools.dataset import Dataset
        >>> from dessia_common.models import all_cars_wi_feat
        >>> filters = [DessiaFilter('weight', '<=', 1650.), DessiaFilter('mpg', '>=', 45.)]
        >>> filters_list = FiltersList(filters, "xor")
        >>> example_list = Dataset(all_cars_wi_feat, name="example")
        >>> filtered_list = example_list.filtering(filters_list)
        >>> print(filtered_list)
        Dataset example: 3 samples, 5 features
        |         Mpg         |    Displacement    |     Horsepower     |       Weight       |    Acceleration    |
        -----------------------------------------------------------------------------------------------------------
        |               35.0  |             0.072  |              69.0  |            1613.0  |              18.0  |
        |               31.0  |             0.076  |              52.0  |            1649.0  |              16.5  |
        |               46.6  |             0.086  |              65.0  |            2110.0  |              17.9  |
        """
        booleans_index = filters_list.get_booleans_index(self.dessia_objects)
        return self._pick_from_boolist(booleans_index)

    def singular_values(self):
        """
        Computes the Singular Values Decomposition (SVD) of self.matrix.

        SVD factorizes self.matrix into two unitary matrices `U` and `Vh`, and a 1-D array `s` of singular values \
            (real, non-negative) such that ``a = U @ S @ Vh``, where S is diagonal such as `s1 > s2 >...> sn`.

        SVD gives indications on the dimensionality of a given matrix thanks to the normalized singular values: they \
            are stored in descending order and their sum is equal to 1. Thus, one can set a threshold value, e.g. \
                `0.95`, and keep only the `r` first normalized singular values which sum is greater than the threshold.

        `r` is the rank of the matrix and gives a good indication on the real dimensionality of the data contained in \
            the current Dataset. `r` is often much smaller than the current dimension of the studied data.
        This indicates that the used features can be combined into less new features, which do not necessarily \
            make sense for engineers.

        More information: https://en.wikipedia.org/wiki/Singular_value_decomposition

        :return:
            **normalized_singular_values**: list of normalized singular values
            **singular_points**: list of points to plot in dimensionality plot. Does not add any information.
        :rtype: Tuple[List[float], List[Dict[str, float]]]
        """
        scaled_data = Dataset._scale_data(npy.array(self.matrix) - npy.mean(self.matrix, axis=0))
        _, singular_values, _ = npy.linalg.svd(npy.array(scaled_data).T, full_matrices=False)
        normalized_singular_values = singular_values / npy.sum(singular_values)

        singular_points = []
        for idx, value in enumerate(normalized_singular_values):
            # TODO (plot_data log_scale 0)
            singular_points.append({'Index of reduced basis vector': idx + 1,
                                    'Singular value': (value if value != 0. else 1e-16)})
        return normalized_singular_values, singular_points

    @staticmethod
    def _scale_data(data_matrix: List[List[float]]):
        scaled_matrix = preprocessing.StandardScaler().fit_transform(data_matrix)
        return [list(map(float, row.tolist())) for row in scaled_matrix]

    def plot_data(self, reference_path: str = "#", **kwargs):
        """ Plot a standard scatter matrix of all attributes in common_attributes and a dimensionality plot. """
        data_list = self._to_samples(reference_path=reference_path)
        if len(self.common_attributes) > 1:
            # Plot a correlation matrix : To develop
            # correlation_matrix = []
            # Dimensionality plot
            dimensionality_plot = self._plot_dimensionality()
            # Scatter Matrix
            scatter_matrix = self._build_multiplot(data_list, self._tooltip_attributes(),
                                                   axis=dimensionality_plot.axis,
                                                   point_style=dimensionality_plot.point_style)
            # Parallel plot
            parallel_plot = self._parallel_plot(data_list)
            return [parallel_plot, scatter_matrix]  # , dimensionality_plot]

        plot_mono_attr = self._histogram_unic_value(0, name_attr=self.common_attributes[0])
        plot_mono_attr.elements = data_list
        return [plot_mono_attr]

    def _build_multiplot(self, data_list: List[Dict[str, float]], tooltip: List[str], **kwargs: Dict[str, Any]):
        subplots = []
        for line in self.common_attributes:
            for idx_col, col in enumerate(self.common_attributes):
                if line == col:
                    subplots.append(self._histogram_unic_value(idx_col, col))
                else:
                    subplots.append(Scatter(x_variable=line, y_variable=col, tooltip=Tooltip(tooltip), **kwargs))

        scatter_matrix = MultiplePlots(plots=subplots, elements=data_list, point_families=self._point_families(),
                                       initial_view_on=True)
        return scatter_matrix

    def _histogram_unic_value(self, idx_col: int, name_attr: str):
        unic_values = set((row_matrix[idx_col] for row_matrix in self.matrix))
        if len(unic_values) == 1:  # TODO (plot_data linspace axis between two same values)
            plot_obj = Scatter(x_variable=name_attr, y_variable=name_attr)
        else:
            plot_obj = Histogram(x_variable=name_attr)
        return plot_obj

    def _tooltip_attributes(self):
        return self.common_attributes

    def _object_to_sample(self, dessia_object: DessiaObject, row: int, reference_path: str = '#'):
        sample_values = {attr: self.matrix[row][col] for col, attr in enumerate(self.common_attributes)}
        reference_path = f"{reference_path}/dessia_objects/{row}"
        name = dessia_object.name if dessia_object.name else f"Sample {row}"
        return Sample(values=sample_values, reference_path=reference_path, name=name)

    def _to_samples(self, reference_path: str = '#'):
        return [self._object_to_sample(dessia_object=dessia_object, row=row, reference_path=reference_path)
                for row, dessia_object in enumerate(self.dessia_objects)]

    def _point_families(self):
        return [PointFamily(GREY, list(range(len(self))))]

    def _parallel_plot(self, data_list: List[Dict[str, float]]):
        return ParallelPlot(elements=data_list, axes=self._parallel_plot_attr(), disposition='vertical')

    def _parallel_plot_attr(self):
        # TODO: Put it in plot_data
        (sorted_r2, sorted_association), constant_attributes = self._get_correlations()
        attribute_series = self._get_attribute_trios(sorted_r2, sorted_association)
        return constant_attributes + self._trios_list_to_parallel_axes(attribute_series)

    def _get_correlations(self):
        r2_scores = []
        association_list = []
        constant_attributes = []
        for idx, attr1 in enumerate(self.common_attributes):
            if len(set(self.attribute_values(attr1))) == 1:
                constant_attributes.append(attr1)
                continue
            for _, attr2 in enumerate(self.common_attributes[idx:]):
                if len(set(self.attribute_values(attr2))) == 1:
                    constant_attributes.append(attr2)
                    continue
                if attr1 != attr2:
                    correlation_matrix = npy.corrcoef(self.attribute_values(attr1),
                                                      self.attribute_values(attr2))
                    correlation_xy = correlation_matrix[0, 1]
                    r2_scores.append(correlation_xy**2)
                    association_list.append([attr1, attr2])

        if len(association_list) == 0:
            association_list = []
            r2_scores = []
            unreal_score = 1.
            for idx, attr1 in enumerate(self.common_attributes):
                for attr2 in self.common_attributes[idx + 1:]:
                    association_list.append([attr1, attr2])
                    r2_scores.append(unreal_score)
                    unreal_score += -1 / 10.
            return map(list, zip(*sorted(zip(r2_scores, association_list))[::-1])), []
        # Returns list of list of associated attributes sorted along their R2 score and constant attributes
        return map(list, zip(*sorted(zip(r2_scores, association_list))[::-1])), list(set(constant_attributes))

    @staticmethod
    def _get_attribute_trios(sorted_r2, sorted_association):
        attribute_series = []
        picked_attr = set()
        set_association = set(sum(sorted_association, []))
        while len(picked_attr) != len(set_association):
            first_association = sorted_association[0]
            attribute_series.append(first_association)
            picked_attr.update(set(first_association))
            del sorted_r2[0], sorted_association[0]

            for idx, _ in enumerate(sorted_r2):
                if any(item in first_association for item in sorted_association[idx]):
                    attribute_series[-1] += sorted_association[idx]
                    picked_attr.update(set(sorted_association[idx]))
                    del sorted_r2[idx], sorted_association[idx]
                    break
        return attribute_series

    def _trios_list_to_parallel_axes(self, attribute_series):
        ordered_attr = []
        for attribute_serie in attribute_series:
            if not any(item in attribute_serie for item in ordered_attr):
                ordered_attr += self._new_attributes_trio(attribute_serie)
            else:
                ordered_attr = self._new_sided_attribute(ordered_attr, attribute_serie)
        return ordered_attr

    @staticmethod
    def _new_attributes_trio(attribute_serie):
        if len(attribute_serie) < 3:
            return attribute_serie
        mid_index = [attribute_serie.count(attr) for attr in attribute_serie].index(2)
        mid_attr = attribute_serie[mid_index]
        remaining_attr = iter(set(attribute_serie).difference({mid_attr}))
        return [next(remaining_attr), mid_attr, next(remaining_attr)]

    @staticmethod
    def _new_sided_attribute(ordered_attr, attribute_serie):
        for side in [0, -1]:
            if ordered_attr[side] in attribute_serie:
                nb_instances = attribute_serie.count(ordered_attr[side])
                for ieme_instance in range(nb_instances):
                    idx_in_serie = (ieme_instance) * 2 + attribute_serie[ieme_instance * 2:].index(ordered_attr[side])
                    # 1 if idx_in_serie = 0, 0 if idx_in_serie = 1, 3 if idx_in_serie = 2, 2 if idx_in_serie = 3
                    idx_attr_to_add = idx_in_serie + 1 - 2 * (idx_in_serie % 2)
                    added_attr = []
                    if attribute_serie[idx_attr_to_add] not in ordered_attr:
                        added_attr = [attribute_serie[idx_attr_to_add]]
                        ordered_attr = (side + 1) * added_attr + ordered_attr + (-1 * side) * added_attr
                        break
        return ordered_attr

    def _plot_dimensionality(self):
        _, singular_points = self.singular_values()

        axis_style = EdgeStyle(line_width=0.5, color_stroke=GREY)
        axis = Axis(nb_points_x=len(singular_points), nb_points_y=len(singular_points), axis_style=axis_style)
        point_style = PointStyle(color_fill=BLUE, color_stroke=BLUE, stroke_width=0.1, size=2, shape='circle')

        dimensionality_plot = Scatter(elements=singular_points, x_variable='Index of reduced basis vector',
                                      y_variable='Singular value', log_scale_y=True, axis=axis, point_style=point_style)
        return dimensionality_plot

    @staticmethod
    def _check_costs(len_data: int, costs: List[List[float]]):
        if len(costs) != len_data:
            if len(costs[0]) == len_data:
                return list(map(list, zip(*costs)))
            raise ValueError(f"costs is length {len(costs)} and the matching Dataset is length {len_data}. " +
                             "They should be the same length.")
        return costs

    @staticmethod
    def pareto_indexes(costs: List[List[float]]):
        """
        Find the Pareto-efficient points.

        :return: A (n_points, ) boolean list, indicating whether each point is Pareto efficient
        """
        is_efficient = npy.ones(len(costs), dtype=bool)
        costs_array = (costs - npy.mean(costs, axis=0)) / npy.std(costs, axis=0)
        for index, cost in enumerate(costs_array):
            if is_efficient[index]:
                # Keep any point with a lower cost
                is_efficient[is_efficient] = npy.any(costs_array[is_efficient] < cost, axis=1)
                # And keep self
                is_efficient[index] = True
        return is_efficient.tolist()

    @staticmethod
    def pareto_frontiers(len_data: int, costs: List[List[float]]):
        """ Experimental method to draw the borders of Pareto domain. """
        checked_costs = Dataset._check_costs(len_data, costs)
        pareto_indexes = Dataset.pareto_indexes(checked_costs)
        pareto_costs = npy.array(list(itertools.compress(checked_costs, pareto_indexes)))

        array_costs = npy.array(checked_costs)
        super_mini = npy.min(array_costs, axis=0)
        pareto_frontiers = []
        for x_dim in range(pareto_costs.shape[1]):
            for y_dim in range(pareto_costs.shape[1]):
                if x_dim != y_dim:
                    frontier_2d = Dataset._pareto_frontier_2d(x_dim, y_dim, pareto_costs,
                                                              npy.max(array_costs[:, x_dim]), super_mini)
                    pareto_frontiers.append(frontier_2d)
        return pareto_frontiers

    @staticmethod
    def _pareto_frontier_2d(x_dim: int, y_dim: int, pareto_costs: List[List[float]], max_x_dim: float,
                            super_mini: List[float]):
        # Experimental
        minidx = npy.argmin(pareto_costs[:, y_dim])
        x_coord = pareto_costs[minidx, x_dim]
        y_coord = pareto_costs[minidx, y_dim]

        new_pareto = pareto_costs[x_coord - pareto_costs[:, x_dim] != 0., :]

        dir_coeffs = (y_coord - new_pareto[:, y_dim]) / (x_coord - new_pareto[:, x_dim])
        dir_coeffs[x_coord == new_pareto[:, x_dim]] = npy.max(dir_coeffs[x_coord != new_pareto[:, x_dim]])

        offsets = y_coord - dir_coeffs * x_coord
        approx_super_mini = dir_coeffs * super_mini[x_dim] + offsets
        chosen_line = npy.argmin(npy.absolute(approx_super_mini - super_mini[y_dim]))

        frontier_2d = npy.array([[super_mini[x_dim], max_x_dim], [approx_super_mini[chosen_line], max_x_dim *
                                                                  dir_coeffs[chosen_line] + offsets[chosen_line]]]).T
        return frontier_2d

    def _compute_costs(self, costs_attributes: List[str]):
        costs = self.sub_matrix(costs_attributes)
        return Dataset._check_costs(len(self), costs)

    def pareto_points(self, costs_attributes: List[str]):
        """
        Find the Pareto-efficient points.

        :param costs_attributes:
            List of columns' attributes on which costs are stored in current Dataset
        :type costs_attributes: List[str]

        :return: a Dataset containing the selected points
        :rtype: Dataset
        """
        checked_costs = self._compute_costs(costs_attributes)
        return self[self.__class__.pareto_indexes(checked_costs)]

    def pareto_sheets(self, costs_attributes: List[str], nb_sheets: int = 1):
        """
        Get successive Pareto sheets (i.e. optimal points in a DOE for already computed costs).

        :param costs_attributes: List of columns' attributes on which costs are stored in current Dataset
        :type costs_attributes: List[str]

        :param nb_sheets: Number of Pareto sheets to pick
        :type nb_sheets: int, `optional`, default to `1`

        :return: The successive Pareto sheets and not selected elements
        :rtype: List[Dataset], Dataset
        """
        checked_costs = self._compute_costs(costs_attributes)
        non_optimal_costs = checked_costs[:]
        non_optimal_points = self.dessia_objects[:]
        pareto_sheets = []
        for idx in range(nb_sheets):
            pareto_sheet = Dataset.pareto_indexes(non_optimal_costs)
            pareto_sheets.append(Dataset(list(itertools.compress(non_optimal_points, pareto_sheet)),
                                         self.name + f'_pareto_{idx}'))
            non_optimal_points = list(itertools.compress(non_optimal_points, map(lambda x: not x, pareto_sheet)))
            non_optimal_costs = list(itertools.compress(non_optimal_costs, map(lambda x: not x, pareto_sheet)))
        return pareto_sheets, Dataset(non_optimal_points, self.name)



"""
Tests on filtering methods with block filter and Dataset.filtering
"""
import json
import importlib_resources
from dessia_common.files import StringFile
from dessia_common.typings import ClassMethodType, MethodType
from dessia_common.tests import Car
from dessia_common.workflow.blocks import ClassMethod, InstantiateModel, Filter, ModelMethod
from dessia_common.core import DessiaFilter, FiltersList
from dessia_common.datatools.dataset import Dataset
from dessia_common.workflow.core import Workflow, Pipe

# Import data
ref = importlib_resources.files("dessia_common").joinpath("models/data/cars.csv")
with ref.open('rb') as csv_cars:
    stream = StringFile.from_stream(csv_cars)

    # =============================================================================================================
    # Filters Workflow
    # =============================================================================================================

    block_0 = ClassMethod(method_type=ClassMethodType(Car, 'from_csv'), name='CSV Cars')
    block_1 = InstantiateModel(model_class=Dataset, name='HList Cars')
    block_2 = ModelMethod(method_type=MethodType(Dataset, 'filtering'), name='Filters HList')
    block_3 = ClassMethod(method_type=ClassMethodType(FiltersList, 'from_filters_list'), name='Filters Hlist')
    block_4 = Filter(filters=[DessiaFilter(attribute='weight', comparison_operator='<=', bound=4000, name='weight'),
                              DessiaFilter(attribute='mpg', comparison_operator='>=', bound=25, name='mpg')],
                     logical_operator='xor', name='Filters on W or MPG')
    blocks = [block_0, block_1, block_2, block_3, block_4]

    pipe_0 = Pipe(block_1.outputs[0], block_2.inputs[0])
    pipe_1 = Pipe(block_3.outputs[0], block_2.inputs[1])
    pipe_2 = Pipe(block_0.outputs[0], block_4.inputs[0])
    pipe_3 = Pipe(block_4.outputs[0], block_1.inputs[0])
    pipes = [pipe_0, pipe_1, pipe_2, pipe_3]

    workflow = Workflow(blocks, pipes, output=block_2.outputs[0], name='Filters demo')

    # Workflow run
    filters = [DessiaFilter('cylinders', ">", 4), DessiaFilter('displacement', ">", 0.25)]
    workflow_run = workflow.run({
        workflow.index(block_0.inputs[0]): stream,
        workflow.index(block_3.inputs[0]): filters,
        workflow.index(block_3.inputs[1]): 'and'})

# Workflow tests
workflow._check_platform()

# JSON TESTS
output_dict = workflow_run.output_value[[0, 3, 9, 11, 25, 44]].to_dict(use_pointers=True)
output_json = json.dumps(output_dict)
output_json_to_dict = json.loads(output_json)
output_jsondict_to_object = Dataset.dict_to_object(output_json_to_dict)

reference_output = {
  "name": "",
  "object_class": "dessia_common.datatools.dataset.Dataset",
  "package_version": "0.9.2.dev320+gecd4609",
  "dessia_objects": [
    {
      "name": "Chevrolet Chevelle Malibu",
      "object_class": "dessia_common.tests.Car",
      "package_version": "0.9.2.dev320+gecd4609",
      "mpg": 18.0,
      "cylinders": 8.0,
      "displacement": 0.307,
      "horsepower": 130.0,
      "weight": 3504.0,
      "acceleration": 12.0,
      "model": 70.0,
      "origin": "US\r"
    },
    {
      "name": "AMC Rebel SST",
      "object_class": "dessia_common.tests.Car",
      "package_version": "0.9.2.dev320+gecd4609",
      "mpg": 16.0,
      "cylinders": 8.0,
      "displacement": 0.304,
      "horsepower": 150.0,
      "weight": 3433.0,
      "acceleration": 12.0,
      "model": 70.0,
      "origin": "US\r"
    },
    {
      "name": "Ford Mustang Boss 302",
      "object_class": "dessia_common.tests.Car",
      "package_version": "0.9.2.dev320+gecd4609",
      "mpg": 0.0,
      "cylinders": 8.0,
      "displacement": 0.302,
      "horsepower": 140.0,
      "weight": 3353.0,
      "acceleration": 8.0,
      "model": 70.0,
      "origin": "US\r"
    },
    {
      "name": "Buick Estate Wagon (sw)",
      "object_class": "dessia_common.tests.Car",
      "package_version": "0.9.2.dev320+gecd4609",
      "mpg": 14.0,
      "cylinders": 8.0,
      "displacement": 0.455,
      "horsepower": 225.0,
      "weight": 3086.0,
      "acceleration": 10.0,
      "model": 70.0,
      "origin": "US\r"
    },
    {
      "name": "AMC Matador",
      "object_class": "dessia_common.tests.Car",
      "package_version": "0.9.2.dev320+gecd4609",
      "mpg": 15.5,
      "cylinders": 8.0,
      "displacement": 0.304,
      "horsepower": 120.0,
      "weight": 3962.0,
      "acceleration": 13.9,
      "model": 76.0,
      "origin": "US\r"
    },
    {
      "name": "Oldsmobile Cutlass Salon Brougham",
      "object_class": "dessia_common.tests.Car",
      "package_version": "0.9.2.dev320+gecd4609",
      "mpg": 23.9,
      "cylinders": 8.0,
      "displacement": 0.26,
      "horsepower": 90.0,
      "weight": 3420.0,
      "acceleration": 22.2,
      "model": 79.0,
      "origin": "US\r"
    }]
}

# Test
assert(output_jsondict_to_object == Dataset.dict_to_object(reference_output))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" Tests for cars data. """

import importlib_resources
from dessia_common.core import DessiaFilter
from dessia_common.tests import Car, CarWithFeatures
from dessia_common.files import StringFile

ref = importlib_resources.files("dessia_common").joinpath("models/data/cars.csv")
with ref.open('rb') as csv_cars:
    stream = StringFile.from_stream(csv_cars)

    filters = [DessiaFilter(attribute="Weight", comparison_operator="lt", bound=4000)]

    # Used models
    all_cars_no_feat = Car.from_csv(stream)
    all_cars_wi_feat = CarWithFeatures.from_csv(stream)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Oct 25 12:28:55 2021

@author: masfaraud
"""

import dessia_common.displays as dcd
import networkx as nx

G = nx.karate_club_graph()

dcd.draw_networkx_graph(G)



"""
Tests for markdowns
"""
from dessia_common.models import all_cars_no_feat
from dessia_common.forms import StandaloneObjectWithDefaultValues
from dessia_common.datatools import HeterogeneousList, CategorizedList
from dessia_common.exports import MarkdownWriter

standalone_object = StandaloneObjectWithDefaultValues()
print(standalone_object.to_markdown())

dataset = HeterogeneousList(all_cars_no_feat)
cars_md = dataset.to_markdown()

clustered_dataset = CategorizedList(all_cars_no_feat)
clustered_dataset.to_markdown()

funky_md = MarkdownWriter().matrix_table([[all_cars_no_feat[0:2], 2, all_cars_no_feat[0]],
                                         [all_cars_no_feat[:5], 3, all_cars_no_feat[10]]],
                                         ['subobject_list', 'object_list', 'carname'])


ref_funky_md = ("| Subobject_list | Object_list | Carname |\n| ------ | ------ | ------ |\n| 2 Cars | 2 | " +
                "Chevrolet Chevelle Malibu |\n| 5 Cars | 3 | Citroen DS-21 Pallas |\n")

assert(all(assertion in cars_md for assertion in ['Model', "0.119", "2625.0"]))
assert(funky_md == ref_funky_md)


writer = MarkdownWriter()
headings = ['Introduction', 'Methodology', 'Results', 'Conclusion']
md = writer.table_of_contents(headings)
for heading in headings:
    assert(heading in md)

print("script 'markdowns.py' has passed")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""

"""

import dessia_common.utils.interpolation as dc_inter

value = 3.4
objects = [1, 2, 3.2, 4.736, 5, 8]
istep = dc_inter.istep_from_value_on_list(objects, value)

inter_value = dc_inter.interpolate_from_istep(objects, istep)

assert value == inter_value



"""
sampling.py package testing.

"""

import json
from typing import Type
from dessia_common.serialization import deserialize_argument
from dessia_common.tests import RandDataD6
from dessia_common.datatools.sampling import ClassSampler
from dessia_common.optimization import FixedAttributeValue, BoundedAttributeValue

sampled_attributes = [BoundedAttributeValue('p_3', 0.1, 0.5, 5),
                      BoundedAttributeValue('p_5', 1500, 5000, 2),
                      BoundedAttributeValue('p_4', 4, 8, 4),
                      BoundedAttributeValue('p_1', 25, 50)]

constant_attributes = [FixedAttributeValue('name', 'sampled_randata'),
                       FixedAttributeValue('p_2', -2.1235),
                       FixedAttributeValue('p_6', 31.1111111)]

randata_sampling = ClassSampler(sampled_class=RandDataD6,
                                sampled_attributes=sampled_attributes,
                                constant_attributes=constant_attributes)

test_hlist = randata_sampling.make_doe('fullfact')
test_hlist = randata_sampling.make_doe(20, 'lhs')
test_hlist = randata_sampling.make_doe(samples=2000, method='montecarlo')
try:
    test_hlist = randata_sampling.make_doe(samples=2000, method='truc')
except NotImplementedError as e:
    assert(e.args[0] == ("Method 'truc' is not implemented in " +
                         "<class 'dessia_common.datatools.sampling.ClassSampler'>._get_doe method."))

json_dict = json.dumps(randata_sampling.to_dict())
decoded_json = json.loads(json_dict)
deserialized_object = randata_sampling.dict_to_object(decoded_json)
randata_sampling._check_platform()

assert(deserialize_argument(Type, "dessia_common.tests.RandDataD6") == RandDataD6)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" Gathers all workflow relative features. """

import ast
import time
import datetime
from functools import cached_property
import io
from typing import List, Union, Type, Any, Dict, Tuple, Optional, TypeVar
import warnings

import humanize
import psutil
import networkx as nx

import dessia_common.errors
from dessia_common.graph import get_column_by_node
from dessia_common.core import DessiaObject
from dessia_common.schemas.core import (get_schema, FAILED_ATTRIBUTE_PARSING, EMPTY_PARSED_ATTRIBUTE,
                                        serialize_annotation, deserialize_annotation, pretty_annotation,
                                        UNDEFINED, Schema, SchemaAttribute)
from dessia_common.utils.types import recursive_type, typematch, is_sequence, is_file_or_file_sequence
from dessia_common.utils.copy import deepcopy_value
from dessia_common.utils.diff import choose_hash
from dessia_common.utils.helpers import prettyname
from dessia_common.typings import JsonSerializable, ViewType
from dessia_common.files import StringFile, BinaryFile
from dessia_common.displays import DisplaySetting
from dessia_common.breakdown import ExtractionError
from dessia_common.errors import SerializationError
from dessia_common.warnings import SerializationWarning
from dessia_common.exports import ExportFormat, MarkdownWriter
import dessia_common.templates
from dessia_common.serialization import (deserialize, serialize_with_pointers, serialize, update_pointers_data,
                                         serialize_dict, add_references, deserialize_argument)
from dessia_common.workflow.utils import ToScriptElement, blocks_to_script, nonblock_variables_to_script


T = TypeVar("T")

VariableAddress = Union[int, Tuple[int, int, int]]


class Variable(DessiaObject):
    """ New version of workflow variable. """

    _eq_is_data_eq = False

    def __init__(self, type_: Type[T] = None, default_value: T = UNDEFINED,
                 name: str = "", label: str = "", position: Tuple[float, float] = (0, 0)):
        self.default_value = default_value
        if self.has_default_value and type_ is None:
            self.type_ = type(default_value)
        else:
            self.type_ = type_
        self.label = label
        self.position = position
        super().__init__(name)

    def to_dict(self, use_pointers: bool = True, memo=None, path: str = '#',
                id_method=True, id_memo=None, **kwargs) -> JsonSerializable:
        """ Customize serialization method in order to handle undefined default value as well as pretty type. """
        dict_ = DessiaObject.base_dict(self)
        dict_.update({"type_": serialize_annotation(self.type_), "position": self.position,
                      "pretty_type": pretty_annotation(self.type_), "label": self.label})
        if self.default_value is not UNDEFINED:
            dict_["default_value"] = serialize(self.default_value)
        return dict_

    @classmethod
    def dict_to_object(cls, dict_: JsonSerializable, **kwargs) -> 'Variable':
        """ Customize serialization method in order to handle undefined default value. """
        default_value = dict_.get("default_value", UNDEFINED)
        label = dict_.get("label", "")  # Backward compatibility < 0.15.0
        return cls(type_=deserialize_annotation(dict_["type_"]), default_value=default_value,
                   name=dict_["name"], label=label, position=tuple(dict_["position"]))

    @property
    def has_default_value(self):
        """ Helper property that indicates if default value should be trusted as such or is undefined. """
        return self.default_value is not UNDEFINED

    @cached_property
    def is_file_related(self) -> bool:
        """ Return whether a variable is of type File given its type_ attribute. """
        schema = get_schema(annotation=self.type_, attribute=SchemaAttribute(self.name))
        return schema.is_file_related

    def copy(self, deep: bool = False, memo=None):
        """
        Copy a Variable.

        :param deep: Deep copy if set to true, shallow copy if false. Defaults to False.
        :param memo: A memo that keeps track of already encountered objects, defaults to None.
        :return: A copy of the object
        """
        if memo is None:
            memo = {}
        if self.has_default_value:
            copied_default_value = deepcopy_value(self.default_value, memo=memo)
        else:
            copied_default_value = UNDEFINED
        return Variable(type_=self.type_, default_value=copied_default_value, name=self.name)


RESULT_VARIABLE_NAME = "_result_name_"


NAME_VARIABLE = Variable(type_=str, name=RESULT_VARIABLE_NAME, label="Result Name")


class Block(DessiaObject):
    """ An Abstract block. Do not instantiate alone. """

    _eq_is_data_eq = False

    def __init__(self, inputs: List[Variable], outputs: List[Variable],
                 position: Tuple[float, float] = (0, 0), name: str = ''):
        self.inputs = inputs
        self.outputs = outputs
        self.position = position
        DessiaObject.__init__(self, name=name)

    def equivalent_hash(self):
        """
        Custom hash of block that doesn't overwrite __hash__ as we do not want to lose python default equality behavior.

        Used by workflow module only.
        """
        return len(self.__class__.__name__)

    def equivalent(self, other):
        """
        Custom equal of block that does not overwrite __eq__ as we do not want to lose python default equality behavior.

        Used by workflow module only.
        """
        return self.__class__.__name__ == other.__class__.__name__

    def _docstring(self):
        """ Base function for sub model docstring computing. """
        block_docstring = {i: EMPTY_PARSED_ATTRIBUTE for i in self.inputs}
        return block_docstring

    def parse_input_doc(self, input_: Variable):
        """ Parse block docstring to get input documentation. """
        try:
            docstring = self._docstring()
            if input_ in docstring:
                return docstring[input_]
        except Exception:  # Broad except to avoid error 500 on doc computing
            return FAILED_ATTRIBUTE_PARSING
        return EMPTY_PARSED_ATTRIBUTE

    def base_script(self) -> str:
        """ Generate a chunk of script that denotes the arguments of a base block. """
        return f"name=\"{self.name}\", position={self.position}"

    def evaluate(self, values, **kwargs):
        """ Not implemented for abstract block class 'evaluate' method. """
        raise NotImplementedError("This method should be implemented in any Block inheriting class.")

    def _to_script(self, prefix: str):
        raise NotImplementedError("This method should be implemented in any Block inheriting class.")

    def is_valid(self, level: str = 'error') -> bool:  # TODO: Change this in further releases
        """ Always return True for now. """
        return True

    def dict_to_inputs(self, dict_: JsonSerializable):
        """
        Enable inputs and outputs overwriting in order to allow input renaming as well as default value persistence.

        If no entry is given in dict, then we have default behavior with blocks generating their own inputs.
        """
        if "inputs" in dict_:
            self.inputs = [Variable.dict_to_object(i) for i in dict_["inputs"]]
        if "outputs" in dict_:
            self.outputs = [Variable.dict_to_object(i) for i in dict_["outputs"]]


class Pipe(DessiaObject):
    """
    Bind two variables of a Workflow.

    :param input_variable: The input variable of the pipe correspond to the start of the arrow, its tail.
    :param output_variable: The output variable of the pipe correspond to the end of the arrow, its hat.
    """

    _eq_is_data_eq = False

    def __init__(self, input_variable: Variable, output_variable: Variable, name: str = ''):
        self.input_variable = input_variable
        self.output_variable = output_variable
        self.memorize = False
        DessiaObject.__init__(self, name=name)

    def to_dict(self, use_pointers=True, memo=None, path: str = '#', id_method=True, id_memo=None,
                **kwargs):
        """ Transform the pipe into a dict. """
        return {'input_variable': self.input_variable, 'output_variable': self.output_variable,
                'memorize': self.memorize}


class WorkflowError(Exception):
    """ Specific WorkflowError Exception. """


class Workflow(Block):
    """
    Class Block of Workflows.

    :param blocks: A List with all the Blocks used by the Workflow.
    :param pipes: A List of Pipe objects.
    :param imposed_variable_values: A dictionary of imposed variable values.
    :param description: A short description that will be displayed on workflow card (frontend).
        Should be shorter than 100 chars
    :param documentation: A long documentation that will be displayed on workflow page (frontend).
        Can use markdown elements.
    :param name: The name of the workflow.
    """

    _standalone_in_db = True
    _eq_is_data_eq = True
    _allowed_methods = ["run", "start_run"]
    _non_serializable_attributes = ["branch_by_display_selector", "branch_by_export_format",
                                    "memorized_pipes", "coordinates", "detached_variables", "variables"]

    def __init__(self, blocks, pipes, output, *, imposed_variable_values=None,
                 detached_variables: List[Variable] = None, description: str = "",
                 documentation: str = "", name: str = ""):
        self.blocks = blocks
        self.pipes = pipes

        if imposed_variable_values is None:
            imposed_variable_values = {}
        self.imposed_variable_values = imposed_variable_values

        self.coordinates = {}

        self.variables = []
        self.nonblock_variables = []
        if detached_variables is None:
            detached_variables = []
        self.detached_variables = detached_variables
        for block in self.blocks:
            self.handle_block(block)

        for pipe in self.pipes:
            self.handle_pipe(pipe)

        self._utd_graph = False

        inputs = [v for v in self.variables if len(nx.ancestors(self.graph, v)) == 0]

        self.description = description
        self.documentation = documentation

        outputs = []
        self.output = output
        if output is not None:
            outputs.append(output)
            self._find_output_block(output)

        self._find_name()

        Block.__init__(self, inputs=inputs, outputs=outputs, name=name)

        self.branch_by_display_selector = self.branch_by_selector(self.display_blocks)
        self.branch_by_export_format = self.branch_by_selector(self.export_blocks)

    @classmethod
    def generate_empty(cls):
        """ Generate an empty workflow (mostly used by frontend to compute an init dict). """
        return cls(blocks=[], pipes=[], output=None)

    @property
    def nodes(self):
        """ Return the list of blocks and non_block_variables (nodes) of the Workflow. """
        return self.blocks + self.nonblock_variables

    @cached_property
    def file_inputs(self):
        """ Get all inputs that are files. """
        return [i for i in self.inputs if i.is_file_related]

    @cached_property
    def has_file_inputs(self) -> bool:
        """ Return True if there is any file input. """
        return any(self.file_inputs)

    @cached_property
    def memorized_pipes(self) -> List[Pipe]:
        """ Get pipes that are memorized. """
        return [p for p in self.pipes if p.memorize]

    def handle_pipe(self, pipe):
        """ Perform some initialization action on a pipe and its variables. """
        upstream_var = pipe.input_variable
        downstream_var = pipe.output_variable
        if upstream_var not in self.variables:
            self.variables.append(upstream_var)
            if upstream_var in self.detached_variables:
                self.detached_variables.remove(upstream_var)
            self.nonblock_variables.append(upstream_var)
        if downstream_var not in self.variables:
            self.variables.append(downstream_var)
            self.nonblock_variables.append(downstream_var)

    def handle_block(self, block):
        """ Perform some initialization action on a block and its variables. """
        if isinstance(block, Workflow):
            # Protecting direct Workflow blocks
            raise ValueError("Using workflow as blocks is forbidden, use WorkflowBlock wrapper instead")

        # Populating variables with block variables
        self.variables.extend(block.inputs)
        self.variables.extend(block.outputs)

        # Memorizing block incoming pipes
        if block in self.display_blocks:
            for i, input_ in enumerate(block.inputs):
                incoming_pipe = self.variable_input_pipe(input_)
                if incoming_pipe and i == block._displayable_input:
                    incoming_pipe.memorize = True

        try:
            self.coordinates[block] = (0, 0)
        except ValueError as err:
            raise ValueError(f"Cannot serialize block {block} ({block.name})") from err

    def _data_hash(self):
        output_hash = hash(self.variable_indices(self.output))
        base_hash = len(self.blocks) + 11 * len(self.pipes) + 23 * len(self.imposed_variable_values) + output_hash
        block_hash = int(sum(b.equivalent_hash() for b in self.blocks) % 1e6)
        return (base_hash + block_hash) % 1000000000

    def _data_eq(self, other_object) -> bool:
        if hash(self) != hash(other_object) or not Block.equivalent(self, other_object):
            return False

        # TODO: temp , reuse graph to handle block order!!!!
        for block1, block2 in zip(self.blocks, other_object.blocks):
            if not block1.equivalent(block2):
                return False

        if not self._equivalent_pipes(other_object):
            return False

        if not self._equivalent_imposed_variables_values(other_object):
            return False
        return True

    def _find_output_block(self, output: Variable):
        found_output = False
        i = 0
        while not found_output and i < len(self.blocks):
            found_output = output in self.blocks[i].outputs
            i += 1
        if not found_output:
            raise WorkflowError("workflow's output is not in any block's outputs")

    def _find_name(self):
        found_name = False
        i = 0
        all_nbvs = self.nonblock_variables + self.detached_variables
        while not found_name and i < len(all_nbvs):
            variable = all_nbvs[i]
            found_name = variable.name == RESULT_VARIABLE_NAME
            i += 1
        if not found_name:
            self.detached_variables.insert(0, NAME_VARIABLE)

    def _equivalent_pipes(self, other_wf) -> bool:
        pipes = []
        other_pipes = []
        for pipe, other_pipe in zip(self.pipes, other_wf.pipes):
            input_index = self.variable_index(pipe.input_variable)
            output_index = self.variable_index(pipe.output_variable)
            pipes.append((input_index, output_index))

            other_input_index = other_wf.variable_index(other_pipe.input_variable)
            other_output_index = other_wf.variable_index(other_pipe.output_variable)
            other_pipes.append((other_input_index, other_output_index))
        return set(pipes) == set(other_pipes)

    def _equivalent_imposed_variables_values(self, other_wf) -> bool:
        ivvs = set()
        other_ivvs = set()
        for key, other_key in zip(self.imposed_variable_values.keys(), other_wf.imposed_variable_values.keys()):
            variable_index = self.variable_index(key)
            ivvs.add((variable_index, self.imposed_variable_values[key]))

            other_variable_index = other_wf.variable_index(other_key)
            other_ivvs.add((other_variable_index, other_wf.imposed_variable_values[other_key]))
        return ivvs == other_ivvs

    def __deepcopy__(self, memo=None):
        """ Return the deep copy. """
        if memo is None:
            memo = {}

        blocks = [deepcopy_value(b, memo) for b in self.blocks]
        output_adress = self.variable_indices(self.output)
        if output_adress is None:
            output = None
        else:
            output_block = blocks[output_adress[0]]
            output = output_block.outputs[output_adress[2]]

        copied_workflow = Workflow(blocks=blocks, pipes=[], output=output, name=self.name)

        pipes = self.copy_pipes(copied_workflow)

        imposed_variable_values = {}
        for variable, value in self.imposed_variable_values.items():
            new_variable = copied_workflow.variable_from_index(self.variable_indices(variable))
            imposed_variable_values[new_variable] = value

        copied_workflow = Workflow(blocks=blocks, pipes=pipes, output=output,
                                   imposed_variable_values=imposed_variable_values, name=self.name)
        return copied_workflow

    def copy_pipe(self, copied_workflow: 'Workflow', pipe: Pipe) -> Pipe:
        """ Copy a single regular pipe. """
        upstream_index = self.variable_indices(pipe.input_variable)
        if self.is_variable_nbv(pipe.input_variable):
            raise dessia_common.errors.CopyError("copy_pipe method cannot handle nonblock-variables. "
                                                 "Please consider using copy_nbv_pipes")
        pipe_upstream = copied_workflow.variable_from_index(upstream_index)
        downstream_index = self.variable_indices(pipe.output_variable)
        pipe_downstream = copied_workflow.variable_from_index(downstream_index)
        return Pipe(pipe_upstream, pipe_downstream)

    def copy_nbv_pipe(self, copied_workflow: 'Workflow', pipe: Pipe, copy_memo: Dict[int, Variable]) -> Pipe:
        """
        Copy a pipe where its upstream variable is a NBV.

        This needs special care because if it is not handled properly, NBVs can duplicate,
        or copied pipes might be unordered.
        """
        nbv = pipe.input_variable
        upstream_index = self.variable_index(nbv)
        if upstream_index in copy_memo:
            copied_variable = copy_memo[upstream_index]
        else:
            copied_variable = nbv.copy()
            copy_memo[upstream_index] = copied_variable
        downstream_index = self.variable_indices(pipe.output_variable)
        pipe_downstream = copied_workflow.variable_from_index(downstream_index)
        return Pipe(copied_variable, pipe_downstream)

    def copy_pipes(self, copied_workflow: 'Workflow') -> List[Pipe]:
        """ Copy all pipes in workflow. """
        copy_memo = {}
        return [self.copy_nbv_pipe(copied_workflow=copied_workflow, pipe=p, copy_memo=copy_memo)
                if self.is_variable_nbv(p.input_variable)
                else self.copy_pipe(copied_workflow=copied_workflow, pipe=p)
                for p in self.pipes]

    def block_selector(self, block: Block) -> str:
        """ Get or create a generic selector for given block. """
        if block in self.display_blocks and block.selector:
            if isinstance(block.selector, ViewType):
                return block.selector.name
            if isinstance(block.selector, str):
                # Backward compatibility dessia_common < 0.14.0
                return block.selector
        return f"{self.selector_name(block)} ({self.blocks.index(block)})"

    def selector_name(self, block: Block) -> str:
        """ Compute name for selector. """
        name = block.name
        if name:
            return name
        if block in self.display_blocks:
            return block.type_
        if block in self.export_blocks:
            return block.extension
        return "Block"

    def branch_by_selector(self, blocks: List[Block]):
        """ Return the corresponding branch to each display or export selector. """
        selector_branches = {}
        for block in blocks:
            branch = self.secondary_branch_blocks(block)
            selector = self.block_selector(block)
            selector_branches[selector] = branch
        return selector_branches

    @property
    def display_blocks(self):
        """ Return list of blocks that can display something (3D, PlotData, Markdown,...). """
        return [b for b in self.blocks if hasattr(b, "_display_settings")]

    @property
    def blocks_display_settings(self) -> List[DisplaySetting]:
        """ Compute all display blocks display_settings. """
        display_settings = []
        for block in self.display_blocks:
            reference_path = "#"
            for i, input_ in enumerate(block.inputs):
                incoming_pipe = self.variable_input_pipe(input_)
                if i == block._displayable_input:
                    reference_path = f"{reference_path}/values/{self.pipes.index(incoming_pipe)}"
            block_index = self.blocks.index(block)
            settings = block._display_settings(block_index=block_index, reference_path=reference_path)
            if settings is not None:
                settings.selector = self.block_selector(block)
                display_settings.append(settings)
        return display_settings

    @staticmethod
    def display_settings(**kwargs) -> List[DisplaySetting]:
        """ Compute the displays settings of the workflow. """
        return [DisplaySetting(selector="Workflow", type_="workflow", method="to_dict", load_by_default=True),
                DisplaySetting(selector="Documentation", type_="markdown", method="to_markdown", load_by_default=True),
                DisplaySetting(selector="Tasks", type_="tasks", method="")]

    @property
    def export_blocks(self):
        """ Return list of blocks that can export something (3D, PlotData, Markdown,...). """
        return [b for b in self.blocks if hasattr(b, "_export_format")]

    @property
    def blocks_export_formats(self):
        """ Compute all export blocks export_formats. """
        export_formats = []
        for block in self.export_blocks:
            block_index = self.blocks.index(block)
            format_ = block._export_format(block_index)
            if format_ is not None:
                format_.selector = self.block_selector(block)
                export_formats.append(format_)
        return export_formats

    def _export_formats(self):
        """ Read block to compute available export formats. """
        export_formats = DessiaObject._export_formats(self)
        script_export = ExportFormat(selector="py", extension="py", method_name="save_script_to_stream", text=True)
        export_formats.append(script_export)
        return export_formats

    def to_markdown(self, **kwargs):
        """ Set workflow documentation as markdown. """
        return self.documentation

    def _docstring(self):
        """ Compute documentation of all blocks. """
        return [b._docstring() for b in self.blocks]

    @property
    def _method_jsonschemas(self):
        """ Compute the run jsonschema (had to be overloaded). """
        warnings.warn("method_jsonschema method is deprecated. Use method_schema instead", DeprecationWarning)
        return self.method_schemas

    @property
    def method_schemas(self):
        """ New support of method schemas. """
        attributes = []
        annotations = {}
        for i, input_ in enumerate(self.inputs + self.detached_variables):
            input_address = str(i)
            annotations[input_address] = input_.type_

            # Title & Description
            description = EMPTY_PARSED_ATTRIBUTE
            title = input_.label
            name = prettyname(input_.name)
            if input_ not in self.nonblock_variables and input_ not in self.detached_variables:
                block = self.block_from_variable(input_)
                description = block.parse_input_doc(input_)
                if not title:
                    title = f"{block.name} - {name}"
            if input_ in self.nonblock_variables and not title:
                title = name

            # Editable and Default values
            editable = input_ not in self.imposed_variable_values
            if input_.has_default_value:
                attributes.append(SchemaAttribute(name=input_address, default_value=input_.default_value, title=title,
                                                  editable=editable, documentation=description))
            else:
                attributes.append(SchemaAttribute(name=input_address, title=title, editable=editable,
                                                  documentation=description))

        schema = Schema(annotations=annotations, attributes=attributes, documentation=self.description)
        schemas = {"run": schema.to_dict(method=True), "start_run": schema.to_dict(method=True, required=[])}
        return schemas

    def to_dict(self, use_pointers=False, memo=None, path="#", id_method=True, id_memo=None, **kwargs):
        """ Compute a dict from the object content. """
        if memo is None:
            memo = {}

        dict_ = Block.to_dict(self, use_pointers=False)

        output = self.variable_indices(self.output)
        dict_.update({"blocks": [b.to_dict(use_pointers=False) for b in self.blocks],
                      "pipes": [self.pipe_variable_indices(p) for p in self.pipes],
                      "output": output,
                      "nonblock_variables": [v.to_dict() for v in self.nonblock_variables + self.detached_variables],
                      "package_mix": self.package_mix()})

        imposed_variable_values = {}
        for variable, value in self.imposed_variable_values.items():
            var_index = self.variable_indices(variable)
            if use_pointers:
                ser_value, memo = serialize_with_pointers(value=value, memo=memo,
                                                          path=f"{path}/imposed_variable_values/{var_index}")
            else:
                ser_value = serialize(value)
            imposed_variable_values[str(var_index)] = ser_value

        dict_.update({"description": self.description, "documentation": self.documentation,
                      "imposed_variable_values": imposed_variable_values})
        return dict_

    @classmethod
    def dict_to_object(cls, dict_: JsonSerializable, **kwargs) -> 'Workflow':
        """ Recompute the object from a dict. """
        pointers_memo = kwargs.get("pointers_memo", None)
        global_dict = kwargs.get("global_dict", None)
        if pointers_memo is None or global_dict is None:
            global_dict, pointers_memo = update_pointers_data(global_dict=global_dict, current_dict=dict_,
                                                              pointers_memo=pointers_memo)

        workflow = initialize_workflow(dict_=dict_, global_dict=global_dict, pointers_memo=pointers_memo)

        imposed_variable_values = {}
        if "imposed_variable_values" in dict_:
            # New format with a dict
            for variable_index_str, serialized_value in dict_["imposed_variable_values"].items():
                value = deserialize(serialized_value, global_dict=global_dict, pointers_memo=pointers_memo)
                variable = workflow.variable_from_index(ast.literal_eval(variable_index_str))
                imposed_variable_values[variable] = value
                variable.default_value = value
        elif "imposed_variable_indices" in dict_:
            for variable_index in dict_["imposed_variable_indices"]:
                variable = workflow.variable_from_index(variable_index)
                imposed_variable_values[variable] = variable.default_value
        else:
            imposed_variable_values = None

        return cls(blocks=workflow.blocks, pipes=workflow.pipes, output=workflow.output,
                   imposed_variable_values=imposed_variable_values, description=dict_.get("description", ""),
                   documentation=dict_.get("documentation", ""), name=dict_["name"])

    def dict_to_arguments(self, dict_: JsonSerializable, method: str, global_dict=None, pointers_memo=None, path='#'):
        """ Process a JSON of arguments and deserialize them. """
        dict_ = {int(k): v for k, v in dict_.items()}  # Serialization set keys as strings
        if method in self._allowed_methods:
            name = None
            arguments_values = {}
            for i, input_ in enumerate(self.inputs):
                overwritten = input_.has_default_value and i in dict_
                imposed = input_ in self.imposed_variable_values
                if (not input_.has_default_value or overwritten) and not imposed:
                    path_value = f"{path}/inputs/{i}"
                    value = deserialize_argument(type_=input_.type_, argument=dict_[i], global_dict=global_dict,
                                                 pointers_memo=pointers_memo, path=path_value)
                    if input_.name == RESULT_VARIABLE_NAME:
                        name = value
                    arguments_values[i] = value
            if name is None and len(self.inputs) in dict_ and isinstance(dict_[len(self.inputs)], str):
                # Hot fixing name not attached
                name = dict_[len(self.inputs)]
            return {"input_values": arguments_values, "name": name}
        raise NotImplementedError(f"Method '{method}' is not allowed for Workflow. Expected 'run' or 'start_run'.")

    def _run_dict(self) -> JsonSerializable:
        dict_ = {}
        for input_ in self.inputs:
            if input_.has_default_value:
                dict_[str(self.variables.index(input_))] = serialize(input_.default_value)
            if input_ in self.imposed_variable_values:
                dict_[str(self.variables.index(input_))] = serialize(self.imposed_variable_values[input_])
        return dict_

    def _start_run_dict(self) -> Dict:
        return {}

    def method_dict(self, method_name: str = None, method_jsonschema=None) -> Dict:
        """ Wrapper method to get dictionaries of run and start_run methods. """
        if method_jsonschema is not None:
            warnings.warn("method_jsonschema argument is deprecated and its use will be removed in a future version."
                          " Please remove it from your function call. Method name is sufficient to get schema",
                          DeprecationWarning)
        if method_name == 'run':
            return self._run_dict()
        if method_name == 'start_run':
            return self._start_run_dict()
        raise WorkflowError(f"Calling method_dict with unknown method_name '{method_name}'")

    def variable_from_index(self, index: VariableAddress):
        """ Index elements are, in order : (Block index : int, Port side (0: input, 1: output), Port index : int). """
        if isinstance(index, int):
            return self.nonblock_variables[index]
        if not index[1]:
            return self.blocks[index[0]].inputs[index[2]]
        return self.blocks[index[0]].outputs[index[2]]

    def _get_graph(self):
        """ Cached property for graph. """
        if not self._utd_graph:
            self._cached_graph = self._graph()
            self._utd_graph = True
        return self._cached_graph

    graph = property(_get_graph)

    def _graph(self):
        """ Compute the networkx graph of the workflow. """
        graph = nx.DiGraph()
        graph.add_nodes_from(self.variables)
        graph.add_nodes_from(self.blocks)
        for block in self.blocks:
            for input_parameter in block.inputs:
                graph.add_edge(input_parameter, block)
            for output_parameter in block.outputs:
                graph.add_edge(block, output_parameter)

        for pipe in self.pipes:
            graph.add_edge(pipe.input_variable, pipe.output_variable)
        return graph

    @property
    def runtime_blocks(self):
        """ Return blocks that are upstream for output. """
        # TODO Check what's happening when output is null (incomplete workflow)
        output_block = self.block_from_variable(self.output)
        output_upstreams = self.upstream_blocks(output_block)
        runtime_blocks = [output_block] + output_upstreams
        i = 0
        while output_upstreams and i <= len(self.blocks):
            block_upstreams = []
            for block in output_upstreams:
                block_upstreams.extend(self.upstream_blocks(block))
            output_upstreams = block_upstreams
            for candidate in block_upstreams:
                if candidate not in runtime_blocks:
                    runtime_blocks.append(candidate)
            i += 1
        return runtime_blocks

    def secondary_branch_blocks(self, block: Block) -> List[Block]:
        """
        Compute the necessary upstream blocks to run a part of a workflow that leads to the given block.

        It stops looking for blocks when it reaches the main branch, and memorize the connected pipe.

        :param block: Block that is the target of the secondary branch
        """
        upstream_blocks = self.upstream_blocks(block)
        branch_blocks = [block]
        i = 0
        candidates = upstream_blocks
        while candidates and i <= len(self.blocks):
            candidates = []
            for upstream_block in upstream_blocks:
                if upstream_block not in self.runtime_blocks and upstream_block not in branch_blocks:
                    branch_blocks.insert(0, upstream_block)
                    candidates.extend(self.upstream_blocks(upstream_block))
            upstream_blocks = candidates
            i += 1
        for branch_block in branch_blocks:
            upstream_blocks = self.upstream_blocks(branch_block)
            for upstream_block in upstream_blocks:
                if upstream_block in self.runtime_blocks:
                    for pipe in self.pipes_between_blocks(upstream_block, branch_block):
                        pipe.memorize = True
        return branch_blocks

    def pipe_from_variable_indices(self, upstream_indices: VariableAddress,
                                   downstream_indices: VariableAddress) -> Pipe:
        """ Get a pipe from the global indices of its attached variables. """
        for pipe in self.pipes:
            if self.variable_indices(pipe.input_variable) == upstream_indices \
                    and self.variable_indices(pipe.output_variable) == downstream_indices:
                return pipe
        msg = f"No pipe has '{upstream_indices}' as upstream variable and '{downstream_indices}' as downstream variable"
        raise ValueError(msg)

    def pipe_variable_indices(self, pipe: Pipe) -> Tuple[VariableAddress, VariableAddress]:
        """ Return the global indices of a pipe's attached variables. """
        return self.variable_indices(pipe.input_variable), self.variable_indices(pipe.output_variable)

    def variable_input_pipe(self, variable: Variable) -> Optional[Pipe]:
        """ Get the incoming pipe for a variable. If variable is not connected, returns None. """
        incoming_pipes = [p for p in self.pipes if p.output_variable == variable]
        if incoming_pipes:  # Inputs can only be connected to one pipe
            incoming_pipe = incoming_pipes[0]
            return incoming_pipe
        return None

    def variable_output_pipes(self, variable: Variable) -> List[Optional[Pipe]]:
        """ Compute all pipes going out a given variable. """
        return [p for p in self.pipes if p.input_variable == variable]

    def pipes_between_blocks(self, upstream_block: Block, downstream_block: Block):
        """ Compute all the pipes linking two blocks. """
        pipes = []
        for outgoing_pipe in self.block_outgoing_pipes(upstream_block):
            if outgoing_pipe is not None and outgoing_pipe in self.block_incoming_pipes(downstream_block):
                pipes.append(outgoing_pipe)
        return pipes

    def block_incoming_pipes(self, block: Block) -> List[Optional[Pipe]]:
        """ Get incoming pipes for every block variable. """
        return [self.variable_input_pipe(i) for i in block.inputs]

    def block_outgoing_pipes(self, block: Block) -> List[Pipe]:
        """ Return all block outgoing pipes. """
        outgoing_pipes = []
        for output in block.outputs:
            outgoing_pipes.extend(self.variable_output_pipes(output))
        return outgoing_pipes

    def upstream_blocks(self, block: Block) -> List[Block]:
        """ Return a list of given block's upstream blocks. """
        # Setting a dict here to foresee a future use. Might be unnecessary
        upstream_variables = {"available": [], "nonblock": [], "wired": []}
        input_upstreams = [self.upstream_variable(i) for i in block.inputs]
        for variable in input_upstreams:
            if variable is None:
                upstream_variables["available"].append(variable)
            elif variable in self.nonblock_variables:
                upstream_variables["nonblock"].append(variable)
            else:
                upstream_variables["wired"].append(variable)
        upstream_blocks = [self.block_from_variable(v) for v in upstream_variables["wired"]]
        return list(set(upstream_blocks))

    def get_upstream_nbv(self, variable: Variable) -> Variable:
        """ If given variable has an upstream nonblock_variable, return it otherwise return given variable itself. """
        if not self.nonblock_variables:
            return variable
        upstream_variable = self.upstream_variable(variable)
        if upstream_variable is not None and upstream_variable in self.nonblock_variables:
            return upstream_variable
        return variable

    def upstream_variable(self, variable: Variable) -> Optional[Variable]:
        """
        Return upstream variable if given variable is connected to a pipe as a pipe output.

        :param variable: Variable to search an upstream for
        """
        incoming_pipe = self.variable_input_pipe(variable)
        if incoming_pipe:
            return incoming_pipe.input_variable
        return None

    def variable_indices(self, variable: Variable) -> Optional[Union[Tuple[int, int, int], int]]:
        """
        Return global address of given variable as a tuple or an int.

        If variable is non block, return index of variable in variables sequence
        Else returns global address (index_block, index, index_port)
        """
        if variable is None:
            return None

        for iblock, block in enumerate(self.blocks):
            if variable in block.inputs:
                return iblock, 0, block.inputs.index(variable)
            if variable in block.outputs:
                return iblock, 1, block.outputs.index(variable)

        upstream_variable = self.get_upstream_nbv(variable)
        if upstream_variable in self.nonblock_variables:
            # Free variable not attached to block
            return self.nonblock_variables.index(upstream_variable)
        raise WorkflowError(f"Something is wrong with variable {variable.name}")

    def is_variable_nbv(self, variable: Variable) -> bool:
        """ Return True if variable does not belong to a block. """
        return isinstance(self.variable_indices(variable), int)

    def block_from_variable(self, variable) -> Block:
        """ Return block of which given variable is attached to. """
        iblock, _, _ = self.variable_indices(variable)
        return self.blocks[iblock]

    def output_disconnected_elements(self):
        """ Return blocks and variables that are not attached to the output. """
        disconnected_elements = []
        ancestors = nx.ancestors(self.graph, self.output)
        for block in self.blocks:
            if block not in ancestors:
                disconnected_elements.append(block)

        for variable in self.nonblock_variables:
            if variable not in ancestors:
                disconnected_elements.append(variable)
        return disconnected_elements

    def index(self, variable):
        """ Deprecated, will be remove in version 0.8.0. """
        warnings.warn("index method is deprecated, use input_index instead", DeprecationWarning)
        return self.input_index(variable)

    def input_index(self, variable: Variable) -> Optional[int]:
        """ If variable is a workflow input, returns its index. """
        upstream_variable = self.get_upstream_nbv(variable)
        if upstream_variable in self.inputs:
            return self.inputs.index(upstream_variable)
        return None

    def variable_index(self, variable: Variable) -> int:
        """ Return variable index in variables sequence. """
        return self.variables.index(variable)

    def block_inputs_global_indices(self, block_index: int) -> List[int]:
        """ Return given block inputs global indices in inputs sequence. """
        block = self.blocks[block_index]
        indices = [self.input_index(i) for i in block.inputs]
        return [i for i in indices if i is not None]

    def match_variables(self, serialize_output: bool = False):
        """
        Run a check for every variable to find its matchable counterparts.

        This means :
            - Variables are compatible workflow-wise
            - Their types are compatible
        """
        variable_match = {}
        for variable in self.variables:
            if variable.type_ is not None:
                vartype = variable.type_
            else:
                continue
            if serialize_output:
                varkey = str(self.variable_indices(variable))
            else:
                varkey = variable
            variable_match[varkey] = []
            for other_variable in self.variables:
                if not self.variable_compatibility(variable, other_variable):
                    continue
                other_vartype = other_variable.type_
                if typematch(vartype, other_vartype):
                    if serialize_output:
                        varval = str(self.variable_indices(other_variable))
                    else:
                        varval = other_variable
                    variable_match[varkey].append(varval)
        return variable_match

    def variable_compatibility(self, variable: Variable, other_variable: Variable) -> bool:
        """
        Check compatibility between variables.

        Two variables are compatible if :
            - They are not equal
            - They don't share the same block
            - They are not input/input or output/output
            - They are typed
        """
        if variable == other_variable:  # If this is the same variable, it is not compatible
            return False

        adress = self.variable_indices(variable)
        other_adress = self.variable_indices(other_variable)

        if variable not in self.nonblock_variables and other_variable not in self.nonblock_variables:
            # If both aren't NBVs we need to check more non-equality elements
            same_block = adress[0] == other_adress[0]
            same_side = adress[1] == other_adress[1]
            if same_block or same_side:
                # A variable cannot be compatible with one on a same block
                # or being the same side (input/input, output/output)
                return False
        # If both are NBVs, non-equality has already been checked
        # If one is NBV and not the other, there is no need to check non-equality

        if variable.type_ is None or other_variable.type_ is None:
            # Variable must be typed to be seen compatible
            return False
        return True

    @property
    def layout_graph(self) -> nx.DiGraph:
        """ Compute graph layout. """
        graph = nx.DiGraph()
        graph.add_nodes_from(self.nodes)  # does not handle detached_variable

        for pipe in self.pipes:
            if pipe.input_variable in self.nonblock_variables:
                input_node = pipe.input_variable
            else:
                input_node = self.block_from_variable(pipe.input_variable)
            output_block = self.block_from_variable(pipe.output_variable)
            graph.add_edge(input_node, output_block)
        return graph

    def graph_columns(self, graph):
        """
        Store nodes of a workflow into a list of nodes indexes.

        :returns: A list of Column Layout where a Column Layout is a list of node_indices
        """
        column_by_node = get_column_by_node(graph)
        nodes_by_column = {}
        for node, column_index in column_by_node.items():
            node_index = self.nodes.index(node)
            nodes_by_column[column_index] = nodes_by_column.get(column_index, []) + [node_index]
        return list(nodes_by_column.values())

    def layout(self):
        """
        Stores a workflow graph layout.

        :returns: A list of Graph Layouts where a GraphLayout is a list of Column Layout
                and Column Layout a list node_indices.
        """
        digraph = self.layout_graph
        graph = digraph.to_undirected()
        connected_components = nx.connected_components(graph)

        return [self.graph_columns(digraph.subgraph(cc)) for cc in list(connected_components)]

    def plot_graph(self):
        """ Plot graph by means of networking and matplotlib. """
        pos = nx.kamada_kawai_layout(self.graph)
        nx.draw_networkx_nodes(self.graph, pos, self.blocks, node_shape='s', node_color='grey')
        nx.draw_networkx_nodes(self.graph, pos, self.variables, node_color='b')
        nx.draw_networkx_nodes(self.graph, pos, self.inputs, node_color='g')
        nx.draw_networkx_nodes(self.graph, pos, self.outputs, node_color='r')
        nx.draw_networkx_edges(self.graph, pos)

        labels = {}  # b: b.function.__name__ for b in self.block}
        for block in self.blocks:
            labels[block] = block.__class__.__name__
            for variable in self.variables:
                labels[variable] = variable.name
        nx.draw_networkx_labels(self.graph, pos, labels)

    def run(self, input_values, verbose=False, progress_callback=lambda x: None, name=None):
        """ Full run of a workflow. Yields a WorkflowRun. """
        log = ""

        state = self.start_run(input_values)
        state.activate_inputs(check_all_inputs=True)

        start_time = time.time()
        start_timestamp = datetime.datetime.now()

        log_msg = "Starting workflow run at {}"
        log_line = log_msg.format(time.strftime("%d/%m/%Y %H:%M:%S UTC", time.gmtime(start_time)))
        log += (log_line + "\n")
        if verbose:
            print(log_line)

        state.continue_run(progress_callback=progress_callback)

        end_time = time.time()
        log_line = f"Workflow terminated in {end_time - start_time} s"

        log += log_line + '\n'
        if verbose:
            print(log_line)

        if not name:
            timestamp = start_timestamp.strftime("%m-%d (%H:%M)")
            name = f"{self.name} @ [{timestamp}]"
        return state.to_workflow_run(name=name)

    def start_run(self, input_values=None, name: str = None):
        """ Partial run of a workflow. Yields a WorkflowState. """
        return WorkflowState(self, input_values=input_values, name=name)

    def jointjs_layout(self, min_horizontal_spacing=300, min_vertical_spacing=200, max_height=800, max_length=1500):
        """ Deprecated workflow layout. Used only in local plot method. """
        coordinates = {}
        elements_by_distance = {}
        if self.output:
            for element in self.nodes:
                distances = []
                paths = nx.all_simple_paths(self.graph, element, self.output)
                for path in paths:
                    distance = 1
                    for path_element in path[1:-1]:
                        if path_element in self.blocks + self.nonblock_variables:
                            distance += 1
                    distances.append(distance)
                try:
                    distance = max(distances)
                except ValueError:
                    distance = 3
                if distance in elements_by_distance:
                    elements_by_distance[distance].append(element)
                else:
                    elements_by_distance[distance] = [element]

        if len(elements_by_distance) != 0:
            max_distance = max(elements_by_distance.keys())
        else:
            max_distance = 3  # TODO: this is an awful quick fix

        horizontal_spacing = max(min_horizontal_spacing, max_length / max_distance)

        for i, distance in enumerate(sorted(elements_by_distance.keys())[::-1]):
            vertical_spacing = min(min_vertical_spacing, max_height / len(elements_by_distance[distance]))
            for j, element in enumerate(elements_by_distance[distance]):
                coordinates[element] = (i * horizontal_spacing, (j + 0.5) * vertical_spacing)
        return coordinates

    def is_valid(self, level: str = "error"):
        """ Tell if the workflow is valid by checking type compatibility of pipes inputs/outputs. """
        for pipe in self.pipes:
            upstream = pipe.input_variable.type_
            downstream = pipe.output_variable.type_
            if upstream.type_ != downstream.type_:
                try:
                    issubclass(upstream.type_.type_, downstream.type_)
                except TypeError as error:  # TODO: need of a real typing check
                    consistent = True
                    if not consistent:
                        raise TypeError(f"Inconsistent pipe type from pipe input '{upstream.name}'"
                                        f"to pipe output '{downstream.name}': "
                                        f"'{upstream.type_}' incompatible with '{downstream.type_}'") from error
        return True

    def package_mix(self) -> Dict[str, float]:
        """ Compute a structure showing percentages of packages used. """
        package_mix = {}
        for block in self.blocks:
            if hasattr(block, 'package_mix'):
                for package_name, fraction in block.package_mix().items():
                    if package_name in package_mix:
                        package_mix[package_name] += fraction
                    else:
                        package_mix[package_name] = fraction

        # Make dimensionless
        fraction_sum = sum(package_mix.values())
        return {pn: f / fraction_sum for pn, f in package_mix.items()}

    def _to_script(self, prefix: str = '') -> ToScriptElement:
        """
        Computes elements for a to_script interpretation.

        :returns: ToSriptElement
        """
        workflow_output_index = self.variable_indices(self.output)
        if workflow_output_index is None:
            raise ValueError("A workflow output must be set")

        imports = []
        imports_as_is = []
        # --- Blocks ---
        blocks_str = blocks_to_script(blocks=self.blocks, prefix=prefix, imports=imports)

        # --- NBVs ---
        nbvs_str = nonblock_variables_to_script(nonblock_variables=self.nonblock_variables, prefix=prefix,
                                                imports=imports, imports_as_is=imports_as_is)

        # --- Pipes ---
        pipes_str = self.pipes_to_script(prefix=prefix, imports=imports)

        # --- Building script ---
        output_name = f"{prefix}block_{workflow_output_index[0]}.outputs[{workflow_output_index[2]}]"

        full_script = f"{blocks_str}\n" \
                      f"{nbvs_str}\n" \
                      f"{pipes_str}\n" \
                      f"{prefix}workflow = " \
                      f"Workflow({prefix}blocks, {prefix}pipes, output={output_name}, documentation=documentation," \
                      f" name=\"{self.name}\")\n"

        self.imposed_variables_to_script(prefix=prefix, full_script=full_script)
        return ToScriptElement(declaration=full_script, imports=imports, imports_as_is=imports_as_is)

    def to_script(self) -> str:
        """ Compute a script representing the workflow. """
        workflow_output_index = self.variable_indices(self.output)
        if workflow_output_index is None:
            raise ValueError("A workflow output must be set")

        self_script = self._to_script()
        self_script.imports.append(self.full_classname)

        script_imports = self_script.imports_to_str()

        return f"{script_imports}\n" \
               f'documentation = """{self.documentation}"""\n\n' \
               f"{self_script.declaration}"

    def pipes_to_script(self, prefix, imports):
        """ Set pipes to script. """
        if len(self.pipes) > 0:
            imports.append(self.pipes[0].full_classname)

        pipes_str = ""
        for ipipe, pipe in enumerate(self.pipes):
            input_index = self.variable_indices(pipe.input_variable)
            if self.is_variable_nbv(pipe.input_variable):  # NBV handling
                input_name = f'{prefix}variable_{input_index}'
            else:
                input_name = f"{prefix}block_{input_index[0]}.outputs[{input_index[2]}]"

            output_index = self.variable_indices(pipe.output_variable)
            if self.is_variable_nbv(pipe.output_variable):  # NBV handling
                output_name = f'{prefix}variable_{output_index}'
            else:
                output_name = f"{prefix}block_{output_index[0]}.inputs[{output_index[2]}]"
            pipes_str += f"{prefix}pipe_{ipipe} = Pipe({input_name}, {output_name})\n"
        pipes_str += f"{prefix}pipes = [{', '.join([prefix + 'pipe_' + str(i) for i in range(len(self.pipes))])}]\n"
        return pipes_str

    def imposed_variables_to_script(self, prefix, full_script):
        """ Set imposed variable values to script. """
        for key, value in self.imposed_variable_values.items():
            variable_indice = self.variable_indices(key)
            if self.is_variable_nbv(key):
                variable_str = variable_indice
            else:
                [block_index, _, variable_index] = variable_indice
                variable_str = f"{prefix}blocks[{block_index}].inputs[{variable_index}]"
            full_script += f"{prefix}workflow.imposed_variable_values[{variable_str}] = {value}\n"

    def save_script_to_stream(self, stream: io.StringIO):
        """ Save the workflow to a python script to a stream. """
        string = self.to_script()
        stream.seek(0)
        stream.write(string)

    def save_script_to_file(self, filename: str):
        """ Save the workflow to a python script to a file on the disk. """
        if not filename.endswith('.py'):
            filename += '.py'
        with open(filename, 'w', encoding='utf-8') as file:
            self.save_script_to_stream(file)

    def evaluate(self, values, **kwargs):
        """ Not implemented Workflow as Block evaluate method. """
        raise NotImplementedError("Method 'evaluate' is not implemented for class Workflow.")


class ExecutionInfo(DessiaObject):
    """ Workflow execution information: start & end date, memory consumption. """

    def __init__(self, start_time: float = None, end_time: float = None,
                 before_block_memory_usage: List[Tuple[Block, int]] = None,
                 after_block_memory_usage: List[Tuple[Block, int]] = None):

        if start_time is None:
            start_time = time.time()

        self.start_time = start_time
        self.end_time = end_time

        if before_block_memory_usage is None:
            before_block_memory_usage = []
        self.before_block_memory_usage = before_block_memory_usage

        if after_block_memory_usage is None:
            after_block_memory_usage = []
        self.after_block_memory_usage = after_block_memory_usage

        DessiaObject.__init__(self, name="")

    @property
    def execution_time(self):
        """ Computes execution time. May return None if end_time is. """
        if self.end_time is None:
            return None
        return self.end_time - self.start_time

    def to_dict(self, use_pointers: bool = True, memo=None, path: str = '#',
                id_method=True, id_memo=None, **kwargs):
        """ Serialize the ExecutionInfo. """
        dict_ = {"start_time": self.start_time, "end_time": self.end_time}
        block_indices = kwargs['block_indices']
        dict_["before_block_memory_usage"] = [(block_indices[b], m) for b, m in self.before_block_memory_usage]
        dict_["after_block_memory_usage"] = [(block_indices[b], m) for b, m in self.after_block_memory_usage]
        return dict_

    @classmethod
    def dict_to_object(cls, dict_: JsonSerializable, **kwargs):
        """ Deserialize the ExecutionInfo. """
        index_to_block = kwargs['index_to_block']
        before_block_memory_usage = [(index_to_block[int(i)], m) for i, m in dict_["before_block_memory_usage"]]
        after_block_memory_usage = [(index_to_block[int(i)], m) for i, m in dict_["after_block_memory_usage"]]
        return cls(start_time=dict_["start_time"], end_time=dict_["end_time"],
                   before_block_memory_usage=before_block_memory_usage,
                   after_block_memory_usage=after_block_memory_usage)

    def to_markdown(self, **kwargs):
        """ Renders to markdown the ExecutionInfo. Requires blocks for clean order. """
        table_content = []
        for block, mem_start in dict(self.after_block_memory_usage):
            mem_end = dict(self.after_block_memory_usage)[block]
            mem_diff = mem_end - mem_start
            table_content.append((block.name, f"{humanize.naturalsize(mem_start)}", f"{humanize.naturalsize(mem_end)}",
                                  f"{humanize.naturalsize(mem_diff)}"))
        writer = MarkdownWriter(print_limit=25, table_limit=None)
        return writer.matrix_table(table_content,
                                   ["Block", "Memory usage at start", "Memory usage at end",
                                    "Memory diff"])


class WorkflowState(DessiaObject):
    """ State of execution of a workflow. """

    _standalone_in_db = True
    _allowed_methods = ['block_evaluation', 'evaluate_next_block', 'continue_run',
                        'evaluate_maximum_blocks', 'add_block_input_values']
    _non_serializable_attributes = ['activated_items']

    def __init__(self, workflow: Workflow, input_values=None, activated_items=None, values=None,
                 output_value=None, log: str = '', execution_info: ExecutionInfo = None, name: str = ''):
        self.workflow = workflow
        if input_values is None:
            input_values = {}
        self.input_values = input_values

        if activated_items is None:
            activated_items = {p: False for p in workflow.pipes}
            activated_items.update({v: False for v in workflow.variables})
            activated_items.update({b: False for b in workflow.blocks})
        self.activated_items = activated_items

        if values is None:
            values = {}
        self.values = values

        if execution_info is None:
            execution_info = ExecutionInfo()
        self.execution_info = execution_info

        self.output_value = output_value
        self.log = log

        self.activate_inputs()
        DessiaObject.__init__(self, name=name)

    def __deepcopy__(self, memo=None):
        if memo is None:
            memo = {}

        workflow = self.workflow.copy(deep=True, memo=memo)
        input_values = deepcopy_value(value=self.input_values, memo=memo)
        values = {}
        for pipe, value in self.values.items():
            variable_indices = self.workflow.pipe_variable_indices(pipe)
            copied_pipe = workflow.pipe_from_variable_indices(*variable_indices)
            values[copied_pipe] = value

        activated_items = {}
        for item, value in self.activated_items.items():
            if isinstance(item, Variable):
                copied_item = workflow.variable_from_index(self.workflow.variable_indices(item))
            elif isinstance(item, Block):
                copied_item = workflow.blocks[self.workflow.blocks.index(item)]
            elif isinstance(item, Pipe):
                copied_item = workflow.pipes[self.workflow.pipes.index(item)]
            else:
                raise ValueError(f"WorkflowState Copy Error : item {item} cannot be activated")
            activated_items[copied_item] = value
        copied_execution_info = self.execution_info.copy(deep=True, memo=memo)
        workflow_state = self.__class__(workflow=workflow, input_values=input_values, activated_items=activated_items,
                                        values=values, output_value=deepcopy_value(value=self.output_value, memo=memo),
                                        log=self.log, execution_info=copied_execution_info, name=self.name)
        return workflow_state

    def _data_hash(self):
        workflow = hash(self.workflow)
        output = choose_hash(self.output_value)
        input_values = sum(i * choose_hash(v) for (i, v) in self.input_values.items())
        values = len(self.values) * 7
        return (workflow + output + input_values + values) % 1000000000

    def _data_eq(self, other_object: 'WorkflowState'):
        if not (self.__class__.__name__ == other_object.__class__.__name__
                and self.progress == other_object.progress
                and self.workflow == other_object.workflow
                and self.output_value == other_object.output_value):
            return False

        for index in set(list(self.input_values.keys()) + list(other_object.input_values.keys())):
            value1 = self.input_values.get(index, None)
            value2 = other_object.input_values.get(index, None)
            if value1 != value2:
                # Rechecking if input is file, in which case we tolerate different values
                if not self.workflow.inputs[index].is_file_related:
                    return False

        for block, other_block in zip(self.workflow.blocks, other_object.workflow.blocks):
            if self.activated_items[block] != other_object.activated_items[other_block]:
                # Check block progress state
                return False
            variables = block.inputs + block.outputs
            other_variables = other_block.inputs + other_block.outputs
            for variable, other_variable in zip(variables, other_variables):
                if self.activated_items[variable] != other_object.activated_items[other_variable]:
                    # Check variables progress state
                    return False

        for pipe, other_pipe in zip(self.workflow.pipes, other_object.workflow.pipes):
            if self.activated_items[pipe] != other_object.activated_items[other_pipe]:
                # Check pipe progress state
                return False
        return True

    @property
    def method_schemas(self):
        """ Empty schemas for WorkflowState because not directly used. """
        return {}

    def to_dict(self, use_pointers: bool = True, memo=None, path: str = '#', id_method=True, id_memo=None,
                **kwargs):
        """ Transform object into a dict. """
        if memo is None:
            memo = {}
        if id_memo is None:
            id_memo = {}

        if use_pointers:
            workflow_dict = self.workflow.to_dict(path=f'{path}/workflow', memo=memo)
        else:
            workflow_dict = self.workflow.to_dict(use_pointers=False)

        execution_info = self.execution_info.to_dict(block_indices={b: i for i, b in enumerate(self.workflow.blocks)})
        dict_ = self.base_dict()
        dict_.update({"execution_info": execution_info, "log": self.log, "workflow": workflow_dict})

        input_values = {}
        for input_number, value in self.input_values.items():
            if self.workflow.inputs[input_number] not in self.workflow.file_inputs:
                if use_pointers:
                    serialized_value, memo = serialize_with_pointers(value=value, memo=memo,
                                                                     path=f"{path}/input_values/{input_number}",
                                                                     id_memo=id_memo)
                else:
                    serialized_value = serialize(value)
                input_values[str(input_number)] = serialized_value

        dict_["input_values"] = input_values

        # Output value: priority for reference before values
        if self.output_value is not None:
            if use_pointers:
                serialized_output_value, memo = serialize_with_pointers(self.output_value, memo=memo,
                                                                        path=f'{path}/output_value',
                                                                        id_memo=id_memo)
            else:
                serialized_output_value = serialize(self.output_value)

            dict_.update({"output_value": serialized_output_value,
                          "output_value_type": recursive_type(self.output_value)})
        # Values
        values = {}
        for pipe, value in self.values.items():
            if not is_file_or_file_sequence(value) and pipe in self.workflow.memorized_pipes:
                pipe_index = self.workflow.pipes.index(pipe)
                if use_pointers:
                    try:
                        serialized_value, memo = serialize_with_pointers(value=value, memo=memo,
                                                                         path=f"{path}/values/{pipe_index}",
                                                                         id_memo=id_memo)
                        values[str(pipe_index)] = serialized_value
                    except SerializationError:
                        warnings.warn(f"unable to serialize {value}, dropping it from workflow state/run values",
                                      SerializationWarning)
                else:
                    values[str(pipe_index)] = serialize(value)
        dict_["values"] = values

        # In the future comment these below and rely only on activated items
        dict_["evaluated_blocks_indices"] = [i for i, b in enumerate(self.workflow.blocks)
                                             if b in self.activated_items and self.activated_items[b]]

        dict_["evaluated_pipes_indices"] = [i for i, p in enumerate(self.workflow.pipes)
                                            if p in self.activated_items and self.activated_items[p]]

        dict_["evaluated_variables_indices"] = [self.workflow.variable_indices(v) for v in self.workflow.variables
                                                if v in self.activated_items and self.activated_items[v]]
        if path == "#":
            add_references(dict_, memo, id_memo)
        return dict_

    @classmethod
    def dict_to_object(cls, dict_: JsonSerializable, global_dict=None, pointers_memo: Dict[str, Any] = None,
                       path: str = "#", **kwargs) -> 'WorkflowState':
        """ Compute Workflow State from given dict. Handles pointers. """
        if pointers_memo is None or global_dict is None:
            global_dict, pointers_memo = update_pointers_data(global_dict=global_dict, current_dict=dict_,
                                                              pointers_memo=pointers_memo)

        workflow = Workflow.dict_to_object(dict_=dict_["workflow"], global_dict=global_dict,
                                           pointers_memo=pointers_memo, path=f"{path}/workflow")

        if "output_value" in dict_:
            output_value = deserialize(dict_["output_value"], global_dict=global_dict,
                                       pointers_memo=pointers_memo, path=f"{path}/output_value")
        else:
            output_value = None

        values = {}
        if "values" in dict_:
            for i, value in dict_["values"].items():
                values[workflow.pipes[int(i)]] = deserialize(value, global_dict=global_dict,
                                                             pointers_memo=pointers_memo, path=f"{path}/values/{i}")

        input_values = {int(i): deserialize(v, global_dict=global_dict, pointers_memo=pointers_memo,
                                            path=f"{path}/input_values/{i}") for i, v in dict_["input_values"].items()}

        activated_items = {b: i in dict_["evaluated_blocks_indices"] for i, b in enumerate(workflow.blocks)}
        activated_items.update({p: i in dict_["evaluated_pipes_indices"] for i, p in enumerate(workflow.pipes)})

        variable_indices = [tuple(i) if is_sequence(i) else i for i in dict_["evaluated_variables_indices"]]
        activated_items.update({v: workflow.variable_indices(v) in variable_indices for v in workflow.variables})

        execution_info = ExecutionInfo.dict_to_object(dict_=dict_["execution_info"],
                                                      index_to_block=dict(enumerate(workflow.blocks)))

        return cls(workflow=workflow, input_values=input_values, activated_items=activated_items, values=values,
                   output_value=output_value, log=dict_["log"], execution_info=execution_info, name=dict_["name"])

    def add_input_value(self, input_index: int, value):
        """ Add a value for given input. """
        self._activate_input(input_=self.workflow.inputs[input_index], value=value)

    def add_several_input_values(self, indices: List[int], values):
        """ Add several values for given inputs. """
        for index in indices:
            input_ = self.workflow.inputs[index]
            if index not in values:
                if self.activated_items[input_] and index in self.input_values:
                    value = self.input_values[index]
                else:
                    msg = f"Value '{input_.name}' of index '{index}' in inputs has no value."
                    if input_.type_ is not None:
                        msg += f" Should be instance of '{input_.type_}'."
                    raise ValueError(msg)
            else:
                value = values[index]
            self.add_input_value(input_index=index, value=value)

    def add_block_input_values(self, block_index: int, values):
        """ Add inputs values for given block. """
        values = {int(k): v for k, v in values.items()}  # Serialization set keys as strings
        indices = self.workflow.block_inputs_global_indices(block_index)
        self.add_several_input_values(indices=indices, values=values)

    def display_settings(self, **kwargs) -> List[DisplaySetting]:
        """ Compute the displays settings of the objects. """
        display_settings = [DisplaySetting(selector="workflow_state", type_="workflow_state", method="state_display")]

        # Displayable blocks
        display_settings.extend(self.workflow.blocks_display_settings)
        return display_settings

    def block_display(self, block_index: int, reference_path: str = "#"):
        """ Compute the display of associated block to use integrate it in the workflow run displays. """
        self.activate_inputs()
        block = self.workflow.blocks[block_index]

        selector = self.workflow.block_selector(block)
        branch = self.workflow.branch_by_display_selector[selector]
        block_args = {}
        for branch_block in branch:
            if branch_block is block:
                argpath = reference_path
            else:
                argpath = "#"
            block_args[branch_block] = {"reference_path": argpath}

        evaluated_blocks = self.evaluate_branch(blocks=branch, block_args=block_args)

        if block not in evaluated_blocks:
            msg = f"Could not reach block at index {block_index}." \
                  f"Has the workflow been run far enough to evaluate this block ?"
            raise WorkflowError(msg)
        return evaluated_blocks[block][0]  # Only one output to an Export Block

    @property
    def progress(self):
        """
        Return the current progress.

        Return a float between 0 (nothing evaluated), to 1 (every computational block evaluated).
        """
        evaluated_blocks = [self.activated_items[b] for b in self.workflow.runtime_blocks]
        progress = sum(evaluated_blocks) / len(evaluated_blocks)
        if progress == 1 and self.execution_info.end_time is None:
            self.execution_info.end_time = time.time()
        return progress

    def block_evaluation(self, block_index: int, progress_callback=lambda x: None) -> bool:
        """ Select a block to evaluate. """
        block = self.workflow.blocks[block_index]
        self.activate_inputs()
        if block in self._activable_blocks():
            nblocks = len(self.workflow.blocks)

            def block_progress_callback(x):
                progress_callback(self.progress + x / nblocks)

            self._evaluate_block(block, progress_callback=block_progress_callback)
            progress_callback(self.progress)
            return True
        return False

    def evaluate_next_block(self, progress_callback=lambda x: None) -> Optional[Block]:
        """ Evaluate a block. """
        self.activate_inputs()
        blocks = self._activable_blocks()
        if blocks:
            block_index = self.workflow.blocks.index(blocks[0])
            self.block_evaluation(block_index, progress_callback=progress_callback)
            return blocks[0]
        return None

    def continue_run(self, progress_callback=lambda x: None, export: bool = False):
        """ Evaluate all possible blocks. """
        self.activate_inputs()

        evaluated_blocks = []
        something_activated = True
        while something_activated:
            something_activated = False
            blocks = [b for b in self.workflow.runtime_blocks if b in self._activable_blocks()]
            for block in blocks:
                evaluated_blocks.append(block)
                block_index = self.workflow.blocks.index(block)
                self.block_evaluation(block_index, progress_callback=progress_callback)
                if not export:
                    progress_callback(self.progress)
                something_activated = True
        return evaluated_blocks

    def evaluate_branch(self, blocks: List[Block], block_args: Dict[Block, Any]):
        """ Evaluate all blocks of a branch, automatically finding the first executable ones. """
        self.activate_inputs()

        if not any((b in self._activable_blocks() for b in blocks)):
            raise WorkflowError("Branch cannot be evaluated because no block has all its inputs activated")

        evaluated_blocks = {}
        i = 0
        while len(evaluated_blocks) != len(blocks) and i <= len(blocks):
            next_blocks = [b for b in blocks if b in self._activable_blocks() and b not in evaluated_blocks]
            for block in next_blocks:
                kwargs = block_args[block]
                output_values = self._evaluate_block(block, **kwargs)
                evaluated_blocks[block] = output_values
            i += 1
        return evaluated_blocks

    def _activate_pipe(self, pipe: Pipe, value):
        """ Set the pipe value and activate its downstream variable. """
        self.values[pipe] = value
        self.activated_items[pipe] = True
        self._activate_variable(variable=pipe.output_variable, value=value)

    def _activate_block(self, block: Block, output_values):
        """ Activate all block outputs. """
        # Unpacking result of evaluation
        output_items = zip(block.outputs, output_values)
        for output, output_value in output_items:
            self._activate_variable(variable=output, value=output_value)
        self.activated_items[block] = True

    def _activate_variable(self, variable: Variable, value):
        """ Activate the given variable with its value and propagate activation to its outgoing pipe. """
        outgoing_pipes = self.workflow.variable_output_pipes(variable)
        if self.workflow.output == variable:
            self.output_value = value
        for outgoing_pipe in outgoing_pipes:
            self._activate_pipe(pipe=outgoing_pipe, value=value)
        self.activated_items[variable] = True

    def _activate_input(self, input_: Variable, value):  # Inputs must always be Typed
        """ Type-check, activate the variable and propagate the value to its pipe. """
        # Type checking
        value_type_check(value, input_.type_)
        input_index = self.workflow.input_index(input_)
        self.input_values[input_index] = value
        self._activate_variable(variable=input_, value=value)
        downstream_pipes = self.workflow.variable_output_pipes(input_)
        for pipe in downstream_pipes:
            self._activate_pipe(pipe=pipe, value=value)

    def _activable_blocks(self):
        """
        Returns a list of all blocks that can be activated.

        Blocks that can be activated are blocks that have all inputs ready for evaluation.
        """
        return [b for b in self.workflow.blocks if self._block_activable_by_inputs(b)
                and (not self.activated_items[b] or b not in self.workflow.runtime_blocks)]

    def _block_activable_by_inputs(self, block: Block):
        """Return whether a block has all its inputs active and can be activated."""
        for function_input in block.inputs:
            if not self.activated_items[function_input]:
                return False
        return True

    def _evaluate_block(self, block, progress_callback=lambda x: x, verbose=False, **kwargs):
        """ Evaluate given block. """
        if verbose:
            log_line = f"Evaluating block {block.name}"
            self.log += log_line + '\n'
            if verbose:
                print(log_line)

        local_values = {}
        for input_ in block.inputs:
            incoming_pipe = self.workflow.variable_input_pipe(input_)
            if incoming_pipe is None:
                # Input isn't connected, it's a workflow input
                input_index = self.workflow.input_index(input_)
                value = self.input_values[input_index]
            else:
                value = self.values[incoming_pipe]
            self._activate_variable(variable=input_, value=value)
            local_values[input_] = value

        kwargs['progress_callback'] = progress_callback
        self.execution_info.before_block_memory_usage.append((block, int(psutil.Process().memory_info().vms)))
        output_values = block.evaluate(local_values, **kwargs)
        self.execution_info.after_block_memory_usage.append((block, int(psutil.Process().memory_info().vms)))
        self._activate_block(block=block, output_values=output_values)

        # Updating progress
        if progress_callback is not None:
            progress_callback(self.progress)
        return output_values

    def activate_inputs(self, check_all_inputs=False):
        """ Return whether all inputs are activated or not. """
        # Input activation
        for index, variable in enumerate(self.workflow.inputs):
            if index in self.input_values:
                self._activate_input(input_=variable, value=self.input_values[index])
            elif variable in self.workflow.imposed_variable_values:
                self._activate_input(input_=variable, value=self.workflow.imposed_variable_values[variable])
            elif variable.has_default_value:
                self._activate_input(input_=variable, value=variable.default_value)
            elif check_all_inputs:
                msg = f"Value {variable.name} of index {index} in inputs has no value"
                if variable.type_ is not None:
                    msg += f": should be instance of {variable.type_}"
                raise ValueError(msg)

    def to_workflow_run(self, name: str = ""):
        """ Return a WorkflowRun if state is complete. """
        if self.progress == 1:
            values = {p: self.values[p] for p in self.workflow.pipes if p in self.values}
            return WorkflowRun(workflow=self.workflow, input_values=self.input_values, output_value=self.output_value,
                               values=values, activated_items=self.activated_items, log=self.log,
                               execution_info=self.execution_info, name=name)
        raise ValueError('Workflow not completed')

    def _export_formats(self):
        """ Read block to compute available export formats. """
        export_formats = DessiaObject._export_formats(self)

        # Exportable Blocks
        export_formats.extend(self.workflow.blocks_export_formats)
        return export_formats

    def export_format_from_selector(self, selector: str):
        """ Get Workflow State format from given selector. """
        for export_format in self.workflow.blocks_export_formats:
            if export_format["selector"] == selector:
                return export_format
        raise ValueError(f"No block defines an export with the selector '{selector}'")

    def export(self, stream: Union[BinaryFile, StringFile], block_index: int):
        """ Perform export. """
        block = self.workflow.blocks[block_index]
        selector = self.workflow.block_selector(block)
        branch = self.workflow.branch_by_export_format[selector]
        block_args = {b: {} for b in branch}
        evaluated_blocks = self.evaluate_branch(blocks=branch, block_args=block_args)
        if block not in evaluated_blocks:
            msg = f"Could not reach block at index {block_index}." \
                  f"Has the workflow been ran far enough to evaluate this block ?"
            raise WorkflowError(msg)
        export_stream = evaluated_blocks[block][0]  # Only one output to an Export Block
        if isinstance(stream, StringFile):
            stream.write(export_stream.getvalue())
        if isinstance(stream, BinaryFile):
            stream.write(export_stream.getbuffer())
        stream.filename = export_stream.filename
        return export_stream

    def to_markdown(self, **kwargs) -> str:
        """ Render to markdown. """
        template = dessia_common.templates.workflow_state_markdown_template
        execution_info = self.execution_info.to_markdown(self.workflow.blocks)
        return template.substitute(name=self.name, class_=self.__class__.__name__, progress=100 * self.progress,
                                   workflow_name=self.workflow.name, execution_info=execution_info)


class WorkflowRun(WorkflowState):
    """ Completed state of a workflow. """

    _allowed_methods = []

    def __init__(self, workflow: Workflow, input_values, output_value, values,
                 activated_items: Dict[Union[Pipe, Block, Variable], bool],
                 log: str = "", execution_info: ExecutionInfo = None, name: str = ""):
        filtered_values = {p: values[p] for p in workflow.memorized_pipes if p in values}
        WorkflowState.__init__(self, workflow=workflow, input_values=input_values,
                               activated_items=activated_items, values=filtered_values,
                               output_value=output_value, log=log, execution_info=execution_info, name=name)

    def to_dict(self, use_pointers: bool = True, memo=None, path: str = '#', id_method=True, id_memo=None,
                **kwargs):
        """ Add variable values to super WorkflowState dict. """
        if memo is None:
            memo = {}  # To make sure we have the good ref for next steps
        return WorkflowState.to_dict(self, use_pointers=use_pointers, memo=memo, path=path,
                                     id_method=id_method, id_memo=id_memo)

    def _get_from_path(self, path: str):
        """
        Extract sub-object at given path. Tries the generic function, then applies specific cases if it fails.

        Returns found object
        """
        try:
            return DessiaObject._get_from_path(self, path)
        except ExtractionError:
            segments = path.split("/")
            first_segment = segments[1]
            if first_segment == "values" and len(segments) >= 3:
                pipe_index = int(segments[2])
                pipe = self.workflow.pipes[pipe_index]
                value = self.values[pipe]
                if len(segments) > 3:
                    return DessiaObject._get_from_path(value, f"#/{'/'.join(segments[3:])}")
                return value
        raise NotImplementedError(f"WorkflowRun : Specific object from path method is not defined for path '{path}'")

    def dict_to_arguments(self, dict_: JsonSerializable, method: str):
        """ Compute run method's arguments from serialized ones. """
        if method in self._allowed_methods:
            return self.workflow.dict_to_arguments(dict_=dict_, method='run')
        raise NotImplementedError(f"Method '{method}' not in WorkflowRun allowed methods")

    def display_settings(self, **kwargs) -> List[DisplaySetting]:
        """
        Compute WorkflowRun display settings.

        Concatenate WorkflowState display_settings and inserting Workflow ones.
        """
        workflow_settings = self.workflow.display_settings()
        block_settings = self.workflow.blocks_display_settings
        displays_by_default = [s.load_by_default for s in block_settings]

        workflow_settings_to_keep = []
        for settings in workflow_settings:
            # Update workflow settings
            settings.compose("workflow")

            if settings.selector == "Workflow":
                settings.load_by_default = False

            if settings.selector == "Documentation":
                settings.load_by_default = not any(displays_by_default)

            if settings.selector != "Tasks":
                workflow_settings_to_keep.append(settings)
        return workflow_settings_to_keep + block_settings

    def method_dict(self, method_name: str = None, method_jsonschema=None):
        """ Get run again default dict. """
        warnings.warn("WorkflowRun's method 'method_dict' is not supported anymore.", DeprecationWarning)
        if method_jsonschema is not None:
            warnings.warn("method_jsonschema argument is deprecated and its use will be removed in a future version."
                          " Please remove it from your function call. Method name is sufficient to get schema",
                          DeprecationWarning)
        if method_name is not None and method_name == 'run_again':
            return serialize_dict(self.input_values)
        raise WorkflowError(f"Calling method_dict with unknown method_name '{method_name}'")

    def run_again(self, input_values, progress_callback=None, name=None):
        """ Execute workflow again with given inputs. """
        return self.workflow.run(input_values=input_values, verbose=False,
                                 progress_callback=progress_callback, name=name)

    @property
    def _method_jsonschemas(self):
        """ Compute the run jsonschema (had to be overloaded). """
        warnings.warn("method_jsonschema method is deprecated. Use method_schema instead", DeprecationWarning)
        return self.method_schemas

    @property
    def method_schemas(self):
        """ Copy old method_jsonschema behavior. Probably to be refactored. """
        warnings.warn("WorkflowRun's method 'method_schemas' is not supported anymore.", DeprecationWarning)
        schemas = {"run_again": self.workflow.method_schemas.pop('run')}
        schemas["run_again"].update({"classes": ["dessia_common.workflow.core.WorkflowRun"], "required": []})
        return schemas

    def to_markdown(self, **kwargs) -> str:
        """ Render to markdown the WorkflowRun. """
        template = dessia_common.templates.workflow_run_markdown_template
        writer = MarkdownWriter(print_limit=25, table_limit=None)

        if is_sequence(self.output_value):
            output_table = f"Output is a sequence of {len(self.output_value)} elements"
        else:
            output_table = writer.object_table(self.output_value)

        execution_info = self.execution_info.to_markdown(blocks=self.workflow.blocks)
        return template.substitute(name=self.name, workflow_name=self.workflow.name,
                                   output_table=output_table, execution_info=execution_info)


def initialize_workflow(dict_, global_dict, pointers_memo) -> Workflow:
    """ Generate blocks, pipes, detached_variables and output from a serialized state. """
    blocks = [deserialize(serialized_element=d, global_dict=global_dict, pointers_memo=pointers_memo)
              for d in dict_["blocks"]]
    if 'nonblock_variables' in dict_:
        nonblock_variables = [deserialize(serialized_element=d, global_dict=global_dict, pointers_memo=pointers_memo)
                              for d in dict_['nonblock_variables']]
    else:
        nonblock_variables = []

    connected_nbvs = {v: False for v in nonblock_variables}
    pipes = deserialize_pipes(pipes_dict=dict_['pipes'], blocks=blocks, nonblock_variables=nonblock_variables,
                              connected_nbvs=connected_nbvs)

    if dict_['output'] is not None:
        output = blocks[dict_['output'][0]].outputs[dict_['output'][2]]
    else:
        output = None
    detached_variables = [v for v, is_connected in connected_nbvs.items() if not is_connected]
    return Workflow(blocks=blocks, pipes=pipes, output=output, detached_variables=detached_variables)


def deserialize_pipes(pipes_dict, blocks, nonblock_variables, connected_nbvs):
    """ Generate all pipes from a dict. """
    pipes = []
    for source, target in pipes_dict:
        if isinstance(source, int):
            variable1 = nonblock_variables[source]
            connected_nbvs[variable1] = True
        else:
            ib1, _, ip1 = source
            variable1 = blocks[ib1].outputs[ip1]

        if isinstance(target, int):
            variable2 = nonblock_variables[target]
            connected_nbvs[variable2] = True
        else:
            ib2, _, ip2 = target
            variable2 = blocks[ib2].inputs[ip2]

        pipes.append(Pipe(variable1, variable2))
    return pipes


def value_type_check(value, type_):
    """
    Type propagation.

    Check if the value as the specified type.
    """
    if type_ is None:
        return False
    try:  # TODO: Sub-scripted generics cannot be used...
        if not isinstance(value, type_):
            return False
    except TypeError:
        pass
    return True



""" Templates for dessia_common. """
from string import Template

dessia_object_markdown_template = Template('''# Object $name of class $class_

This is a markdown file https://www.markdownguide.org/cheat-sheet/

The good practice is to create a string python template and move the template to another python module
(like templates.py) to avoid mix python code and markdown, as python syntax conflicts with markdown

You can substitute values with object attributes like the name of the object: $name

# Attributes

Object $name of class $class_ has the following attributes:

$table

''')


visjs_template = Template('''<html>
        <head>
        <script src="https://cdnjs.cloudflare.com/ajax/libs/vis/4.21.0/vis.min.js"></script>
        <link href="https://cdnjs.cloudflare.com/ajax/libs/vis/4.21.0/vis.min.css" rel="stylesheet" />
        <meta charset="utf-8"/>
        <style type="text/css">
            #mynetwork {
                border: 1px solid lightgray;
            }
        </style>
        <title>$name</title>
    </head>
    <body>
    <div id="mynetwork"></div>

    <script type="text/javascript">
    var nodes = new vis.DataSet($nodes);

var edges = new vis.DataSet($edges);


// create a network
    var container = document.getElementById('mynetwork');

    // provide the data in the vis format
    var data = {
        nodes: nodes,
        edges: edges
    };

    scale=function (min,max,total,value) {
      if (max === min) {
        return 0.5;
      }
      else {
        var scale = 1 / (max);
        return Math.max(0,(value )*scale);
      }
    }

    var options = {edges: {scaling: {'customScalingFunction': scale}}};

    // initialize your network!
    var network = new vis.Network(container, data, options);
</script>
</body>
</html>
''')


dataset_markdown_template = Template('''# Dataset $name of $element_details:

$table

## Information:

    This is a standard markdown: https://www.markdownguide.org/cheat-sheet/

    The good practice is to create a string python template and move the template to another python module
    (like templates.py) to avoid mix python code and markdown, as python syntax conflicts with markdown

    You can substitute values with object attributes like the name of the object: $name

''')


workflow_state_markdown_template = Template('''# WorkflowState $name

progress $progress% on workflow $workflow_name

# Execution informations

$execution_info

''')

workflow_run_markdown_template = Template('''# WorkflowRun $name

Run of workflow $workflow_name

# Output

$output_table

# Execution informations

$execution_info

''')



import unittest

    
class BackendBreakingChangeTest(unittest.TestCase):
    def test_import_is_working(self):
        """Basic unittest to make sure backend import of DC is working"""
        try:
            from dessia_common import __version__
            from dessia_common.core import DessiaObject, inspect_arguments, stringify_dict_keys
            from dessia_common.errors import DeepAttributeError
            from dessia_common.files import BinaryFile, StringFile
            from dessia_common.utils.jsonschema import default_dict
            from dessia_common.utils.serialization import serialize, serialize_with_pointers
            from dessia_common.utils.types import TYPING_EQUIVALENCES, is_bson_valid, is_jsonable, is_sequence, serialize_typing
            from dessia_common.workflow.core import WorkflowState        
        except:
            self.fail("Breaking changes for backend detected. Please create a backend ticket with changes for fixing this UT.")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" Graph helpers. """
import networkx as nx


def explore_tree_from_leaves(graph):
    """ Explore a tree from leaves to top. """
    if not nx.is_directed_acyclic_graph(graph):
        raise NotImplementedError('Cycles in jsonpointers not handled')

    return list(nx.topological_sort(graph.reverse()))


def cut_tree_final_branches(graph: nx.DiGraph):
    """ Cut final branches of a decision tree. """  # TODO: Check this
    number_node_removed = 1
    while number_node_removed:
        nodes_to_delete = []
        # number_node_removed = 0
        for node in graph.nodes():
            if graph.out_degree(node) == 0:
                # It is a leaf
                degree_in = graph.in_degree(node)
                current_node = node
                while degree_in == 1:
                    nodes_to_delete.append(current_node)
                    current_node = graph.predecessors(current_node)
                    degree_in = graph.in_degree(current_node)
        number_node_removed = len(nodes_to_delete)
        if number_node_removed:
            new_nodes = [n for n in graph.nodes if n not in nodes_to_delete]
            graph = nx.subgraph(graph, new_nodes)

    return graph


def extract_region(networkx_graph: nx.Graph, nodes, distance: int = 5):
    """ Extract all nodes that are at distance from node. """
    region_nodes = nodes[:]
    for node in nodes:
        if node not in networkx_graph:
            raise ValueError(f'node {node} is not in graph')
        succs = list(nx.dfs_successors(networkx_graph, source=node, depth_limit=distance))
        preds = list(nx.dfs_predecessors(networkx_graph, source=node, depth_limit=distance))
        ancs = list(nx.ancestors(networkx_graph, source=node))
        print(f'found {len(succs)} successors and {len(preds)} predecessors {len(ancs)} ancestors')
        region_nodes.extend(succs)
        region_nodes.extend(preds)
        region_nodes.extend(ancs)

    return networkx_graph.subgraph(region_nodes)


def get_longest_path(graph: nx.DiGraph, begin, end):
    """ Get longest path of graph. """
    return list(nx.shortest_simple_paths(graph, begin, end))[-1]


def get_paths_from_to(graph: nx.DiGraph, origins, destinations):
    """
    Get all paths from origins to destinations.

    :param graph: a digraph
    :param origins: a list of nodes in graph
    :param destinations: a list of nodes in graph
    :returns: a list of the longest paths where the beginning is in origins and the end is in destinations
    """
    paths = []
    for origin in origins:
        for destination in destinations:
            try:
                paths.append(get_longest_path(graph, origin, destination))
            except nx.exception.NetworkXNoPath:
                continue
    return paths


def get_column_by_node(graph: nx.DiGraph):
    """ Get column values corresponding to nodes and store it in a dict containing the column_index of each node. """
    longest_path = nx.dag_longest_path(graph)
    end_of_path = longest_path[-1]

    columns = {}

    for node in longest_path:
        columns[node] = longest_path.index(node)

    for node in [n for n in graph.nodes if n not in longest_path]:
        try:
            path = get_longest_path(graph, node, end_of_path)
            columns[node] = len(longest_path) - len(path)
        except nx.exception.NetworkXNoPath:
            # There is no path from node to end_of_path
            for path in get_paths_from_to(graph, list(graph.nodes), [node]):
                column_index = columns[path[0]] + len(path) - 1
                if node not in columns or columns[node] < column_index:
                    columns[node] = column_index

    return columns
